"""
Module responsible for handling Flit python environments.


Author: Tom Fleet
Created: 26/12/2021
"""

from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path

from pytoil.environments.virtualenv import Venv
from pytoil.exceptions import FlitNotInstalledError

FLIT = shutil.which("flit")


class Flit(Venv):
    def __init__(self, root: Path, flit: str | None = FLIT) -> None:
        self.root = root
        self.flit = flit
        super().__init__(root)

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r}, flit={self.flit!r})"

    __slots__ = ("root", "flit")

    @property
    def name(self) -> str:
        return "flit"

    def install_self(self, silent: bool = False) -> None:
        """
        Installs a flit based project.

        Args:
            silent (bool, optional): Whether to discard or display output.
                Defaults to False.
        """
        if not self.flit:
            raise FlitNotInstalledError

        # Unlike poetry, conda etc. flit does not make it's own virtual environment
        # we must make one here before installing the project
        if not self.exists():
            self.create()

        subprocess.run(
            [
                self.flit,
                "install",
                "--deps",
                "develop",
                "--symlink",
                "--python",
                f"{self.executable}",
            ],
            cwd=self.project_path,
            stdout=subprocess.DEVNULL if silent else sys.stdout,
            stderr=subprocess.DEVNULL if silent else sys.stderr,
        )



from __future__ import annotations

import sys
import tempfile
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.exceptions import CargoNotInstalledError
from pytoil.starters import RustStarter


def test_go_starter_init():
    starter = RustStarter(path=Path("somewhere"), name="testyrust", cargo="notcargo")

    assert starter.path == Path("somewhere")
    assert starter.name == "testyrust"
    assert starter.root == Path("somewhere").joinpath("testyrust").resolve()
    assert starter.files == [
        Path("somewhere").joinpath("testyrust").resolve().joinpath("README.md"),
    ]


def test_generate_raises_if_cargo_not_installed():
    starter = RustStarter(path=Path("somewhere"), name="testyrust", cargo=None)

    with pytest.raises(CargoNotInstalledError):
        starter.generate()


def test_rust_starter_generate(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmpdir:
        starter = RustStarter(path=Path(tmpdir), name="temprust", cargo="notcargo")

        mock_cargo_init = mocker.patch(
            "pytoil.starters.rust.subprocess.run", autospec=True
        )

        starter.generate()

        mock_cargo_init.assert_called_once_with(
            ["notcargo", "init", "--vcs", "none"],
            cwd=starter.root,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        for file in starter.files:
            assert file.exists()

        with open(starter.root.joinpath("README.md")) as readme:
            readme_content = readme.read()

        assert readme_content == "# temprust\n"



"""
The pytoil info command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import click
from rich import box
from rich.console import Console
from rich.table import Table

from pytoil.api import API
from pytoil.cli.printer import printer
from pytoil.config import Config
from pytoil.exceptions import RepoNotFoundError
from pytoil.repo import Repo


@click.command()
@click.argument("project", nargs=1)
@click.pass_obj
def info(config: Config, project: str) -> None:
    """
    Get useful info for a project.

    Given a project name (can be local or remote), 'info' will return a summary
    description of the project.

    If the project is on GitHub, info will prefer getting information from the GitHub
    API as this is more detailed.

    If the project is local only, some information is extracted from the operating
    system about the project.

    Examples:

    $ pytoil info my_project
    """
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        info = repo.info(api)
    except RepoNotFoundError:
        printer.error(
            f"{project!r} not found locally or on GitHub. Was it a typo?", exits=1
        )
    else:
        table = Table(box=box.SIMPLE)
        table.add_column("Key", style="cyan", justify="right")
        table.add_column("Value", justify="left")

        for key, val in info.items():
            table.add_row(f"{key}:", str(val))

        console = Console()
        console.print(table)



from __future__ import annotations

import sys
from pathlib import Path

from pytest_mock import MockerFixture

from pytoil.editor import launch


def test_launch(mocker: MockerFixture):
    mock = mocker.patch("pytoil.editor.editor.subprocess.run", autospec=True)

    launch(path=Path("somewhere"), bin="/path/to/editor")

    mock.assert_called_once_with(
        ["/path/to/editor", Path("somewhere")], stdout=sys.stdout, stderr=sys.stderr
    )



from __future__ import annotations

from pathlib import Path

import pytest
import rtoml
import yaml


@pytest.fixture
def fake_get_repo_names_response():
    """
    Response snippet for the get_repo_names GraphQL
    query.
    """

    return {
        "data": {
            "user": {
                "repositories": {
                    "nodes": [
                        {"name": "dingle"},
                        {"name": "dongle"},
                        {"name": "dangle"},
                        {"name": "a_cool_project"},
                        {"name": "another"},
                        {"name": "yetanother"},
                        {"name": "hello"},
                    ]
                }
            }
        }
    }


@pytest.fixture
def fake_get_repos_response():
    """
    Response snippet for the get_repos GraphQL
    query.
    """
    return {
        "data": {
            "user": {
                "repositories": {
                    "nodes": [
                        {
                            "name": "advent_of_code_2020",
                            "description": "Retroactively doing AOC2020 in Go.",
                            "createdAt": "2022-01-05T16:54:03Z",
                            "pushedAt": "2022-01-09T06:55:32Z",
                            "diskUsage": 45,
                        },
                        {
                            "name": "advent_of_code_2021",
                            "description": "My code for AOC 2021",
                            "createdAt": "2021-11-30T12:01:22Z",
                            "pushedAt": "2021-12-19T15:10:07Z",
                            "diskUsage": 151,
                        },
                        {
                            "name": "aircraft_crashes",
                            "description": "Analysis of aircraft crash data.",
                            "createdAt": "2021-01-02T19:34:15Z",
                            "pushedAt": "2021-01-20T10:35:57Z",
                            "diskUsage": 2062,
                        },
                        {
                            "name": "cookie_pypackage",
                            "description": (
                                "My own version of the Cookiecutter pypackage template"
                            ),
                            "createdAt": "2020-07-04T10:05:36Z",
                            "pushedAt": "2021-12-03T08:45:49Z",
                            "diskUsage": 734,
                        },
                        {
                            "name": "cv",
                            "description": "Repo for my CV, built with JSON Resume.",
                            "createdAt": "2021-10-30T15:11:49Z",
                            "pushedAt": "2022-01-10T19:50:24Z",
                            "diskUsage": 145,
                        },
                        {
                            "name": "eu_energy_analysis",
                            "description": "Analysis of the EU Open Power System Data.",
                            "createdAt": "2020-12-13T10:50:35Z",
                            "pushedAt": "2020-12-24T11:12:34Z",
                            "diskUsage": 1834,
                        },
                        {
                            "name": "FollowTheProcess",
                            "description": 'My "About Me" Repo',
                            "createdAt": "2020-07-14T16:06:52Z",
                            "pushedAt": "2022-01-10T20:05:47Z",
                            "diskUsage": 14640,
                        },
                        {
                            "name": "followtheprocess.github.io",
                            "description": "Repo for my GitHub pages site.",
                            "createdAt": "2021-02-19T20:16:05Z",
                            "pushedAt": "2021-11-18T19:04:06Z",
                            "diskUsage": 10753,
                        },
                    ]
                }
            }
        }
    }


@pytest.fixture
def fake_get_forks_response():
    """
    Response snippet for the get_forks GraphQL
    query.
    """
    return {
        "data": {
            "user": {
                "repositories": {
                    "nodes": [
                        {
                            "name": "nox",
                            "diskUsage": 5125,
                            "createdAt": "2021-07-01T11:43:36Z",
                            "pushedAt": "2022-01-08T11:00:44Z",
                            "parent": {"nameWithOwner": "theacodes/nox"},
                        },
                        {
                            "name": "python-launcher",
                            "diskUsage": 824,
                            "createdAt": "2021-10-25T18:33:11Z",
                            "pushedAt": "2021-11-09T07:47:23Z",
                            "parent": {"nameWithOwner": "brettcannon/python-launcher"},
                        },
                    ]
                }
            }
        }
    }


@pytest.fixture
def fake_repo_exists_false_response():
    """
    Response snippet for the check_repo_exists GraphQL
    query when the requested repo does not exist.
    """
    return {
        "data": {"repository": None},
        "errors": [
            {
                "type": "NOT_FOUND",
                "path": ["repository"],
                "locations": [{"line": 7, "column": 3}],
                "message": (
                    "Could not resolve to a Repository with the name"
                    " 'FollowTheProcess/dave'."
                ),
            }
        ],
    }


@pytest.fixture
def fake_repo_exists_true_response():
    """
    Response snippet for the check_repo_exists GraphQL
    query when the requested repo exists.
    """
    return {"data": {"repository": {"name": "pytoil"}}}


@pytest.fixture
def fake_repo_info_response():
    """
    Response snippet for the get_repo_info GraphQL
    query.
    """
    return {
        "data": {
            "repository": {
                "name": "pytoil",
                "description": "CLI to automate the development workflow :robot:",
                "createdAt": "2021-02-04T15:05:23Z",
                "pushedAt": "2021-12-27T13:31:53Z",
                "diskUsage": 3153,
                "licenseInfo": {"name": "Apache License 2.0"},
                "primaryLanguage": {"name": "Python"},
            }
        }
    }


@pytest.fixture
def fake_repo_info_response_no_license():
    """
    Response snippet for the get_repo_info GraphQL
    query.
    """
    return {
        "data": {
            "repository": {
                "name": "pytoil",
                "description": "CLI to automate the development workflow :robot:",
                "createdAt": "2021-02-04T15:05:23Z",
                "pushedAt": "2021-12-27T13:31:53Z",
                "diskUsage": 3153,
                "licenseInfo": None,
                "primaryLanguage": {"name": "Python"},
            }
        }
    }


@pytest.fixture
def fake_projects_dir(tmp_path_factory):
    """
    Returns a fake projects directory complete
    with subdirectories mocking development projects.
    """

    project_names: list[str] = [
        "project1",
        "myproject",
        "dingleproject",
        "anotherone",
        ".ishouldnt_show_up",
    ]

    projects_dir: Path = tmp_path_factory.mktemp("projects")

    for project_name in project_names:
        project_path = projects_dir.joinpath(project_name)
        project_path.mkdir()

    return projects_dir


@pytest.fixture
def temp_environment_yml(tmp_path_factory):
    """
    Returns a valid temporary environment.yml file
    """

    env_file: Path = tmp_path_factory.mktemp("temp").joinpath("environment.yml")

    fake_env_info: dict[str, str | list[str]] = {
        "name": "testy",
        "channels": ["defaults", "conda-forge"],
        "dependencies": [
            "python=3",
            "invoke",
            "black",
            "flake8",
            "isort",
            "mypy",
            "rich",
            "numpy",
            "requests",
            "pandas",
        ],
    }

    with open(env_file, mode="w", encoding="utf-8") as f:
        content = yaml.dump(fake_env_info)
        f.write(content)

    return env_file


@pytest.fixture
def bad_temp_environment_yml(tmp_path_factory):
    """
    Returns an invalid temporary environment.yml file
    """

    env_file: Path = tmp_path_factory.mktemp("temp").joinpath("environment.yml")

    # name must be a string
    fake_env_info = {
        "name": 300,
        "channels": ["defaults", "conda-forge"],
        "dependencies": [
            "python>3.7",
            "invoke",
            "black",
            "flake8",
            "isort",
            "mypy",
            "rich",
            "requests",
            "numpy",
            "pandas",
        ],
    }

    with open(env_file, "w") as f:
        content = yaml.dump(fake_env_info)
        f.write(content)

    return env_file


@pytest.fixture
def fake_home_folder_miniconda(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().
    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: Path = tmp_path_factory.mktemp("home")

    miniconda = fake_home.joinpath("miniconda3/envs")
    miniconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_anaconda(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().
    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: Path = tmp_path_factory.mktemp("home")

    anaconda = fake_home.joinpath("anaconda3/envs")
    anaconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_miniforge(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().
    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: Path = tmp_path_factory.mktemp("home")

    anaconda = fake_home.joinpath("miniforge3/envs")
    anaconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_mambaforge(tmp_path_factory):
    """
    Returns a faked $HOME folder. Should be used as the return
    value from pathlib.Path.home().
    Designed to test the auto detection of conda environment
    storage directory.
    """

    fake_home: Path = tmp_path_factory.mktemp("home")

    anaconda = fake_home.joinpath("mambaforge/envs")
    anaconda.mkdir(parents=True)

    return fake_home


@pytest.fixture
def fake_home_folder_no_conda(tmp_path_factory):
    """
    Returns a faked $HOME but without any conda dirs.
    To test whether get_envs_dir raises correctly.
    """

    fake_home: Path = tmp_path_factory.mktemp("home")

    return fake_home


@pytest.fixture
def repo_folder_with_random_existing_files(tmp_path_factory):
    """
    Returns a temporary directory containing a few random files
    to test methods which check for a files existence within
    a certain directory.
    """

    folder: Path = tmp_path_factory.mktemp("myrepo")

    files: list[str] = ["here.txt", "i_exist.yml", "hello.py", "me_too.json"]

    for file in files:
        # Create each file under the myrepo folder
        folder.joinpath(file).touch()

    return folder


@pytest.fixture
def project_with_no_build_system(tmp_path_factory):
    """
    Returns a temporary directory containing a
    pyproject.toml but this file does not specify
    a build-system, which is bad.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    return folder


@pytest.fixture
def project_with_no_build_backend(tmp_path_factory):
    """
    Returns a temporary directory containing a
    pyproject.toml.
    This time we do write a 'build-system' but no
    'build-backend'.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    # Create some fake build-system
    build_system = {
        "build-system": {
            "requires": ["poetry-core>=1.0.0"],
        },
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(build_system, f)

    return folder


@pytest.fixture
def project_with_setuptools_pep621_backend(tmp_path_factory):
    """
    Returns a temporary directory containing a PEP 621
    pyproject.toml configured to use setuptools as the
    build backend.
    """
    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    build_system = {
        "build-system": {
            "requires": [
                "setuptools>=61",
                "wheel>=0.37.1",
            ],
            "build-backend": "setuptools.build_meta",
        }
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(build_system, f)

    return folder


@pytest.fixture
def fake_poetry_project(tmp_path_factory):
    """
    Returns a temporary directory containing a
    valid poetry pyproject.toml file.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    # Create some fake poetry content
    build_system = {
        "build-system": {
            "requires": ["poetry-core>=1.0.0"],
            "build-backend": "poetry.core.masonry.api",
        },
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(build_system, f)

    return folder


@pytest.fixture
def fake_flit_project(tmp_path_factory):
    """
    Returns a temporary directory containing a
    valid flit pyproject.toml file.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    # Create some fake flit content
    build_system = {
        "build-system": {
            "requires": ["flit_core >=2,<4"],
            "build-backend": "flit_core.buildapi",
        },
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(build_system, f)

    return folder


@pytest.fixture
def fake_hatch_project(tmp_path_factory):
    """
    Returns a temporary directory containing a
    valid hatch pyproject.toml file.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    # Create some fake hatch content
    build_system = {
        "build-system": {
            "requires": ["hatchling"],
            "build-backend": "hatchling.build",
        },
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(build_system, f)

    return folder


@pytest.fixture
def fake_pep621_project(tmp_path_factory):
    """
    Returns a temporary directory containing a
    valid pep621 pyproject.toml file.
    """

    # Create the folder and pyproject.toml file
    folder: Path = tmp_path_factory.mktemp("myrepo")
    pyproject_toml = folder.joinpath("pyproject.toml")
    pyproject_toml.touch()

    # Create some fake pep621 content
    content = {
        "build-system": {
            "requires": ["hatchling"],
            "build-backend": "hatchling.build",
        },
        "project": {
            "name": "myrepo",
            "version": "0.1.0",
            "description": "A test project",
            "readme": "README.md",
        },
    }

    with open(pyproject_toml, mode="w", encoding="utf-8") as f:
        rtoml.dump(content, f)

    return folder


@pytest.fixture
def requirements_dev_project(tmp_path_factory):
    """
    Returns a temp directory containing a requirements-dev.txt
    file.
    """

    folder: Path = tmp_path_factory.mktemp("myrepo")
    req_dev_txt = folder.joinpath("requirements-dev.txt")
    req_dev_txt.touch()

    return folder


@pytest.fixture
def requirements_project(tmp_path_factory):
    """
    Returns a temp directory containing a requirements.txt
    file.
    """

    folder: Path = tmp_path_factory.mktemp("myrepo")
    req_txt = folder.joinpath("requirements.txt")
    req_txt.touch()

    return folder



"""
The pytoil gh command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import click
import httpx

from pytoil.api import API
from pytoil.cli import utils
from pytoil.cli.printer import printer
from pytoil.config import Config
from pytoil.repo import Repo


@click.command()
@click.argument("project", nargs=1)
@click.option("-i", "--issues", is_flag=True, help="Go to the issues page.")
@click.option("-p", "--prs", is_flag=True, help="Go to the pull requests page.")
@click.pass_obj
def gh(config: Config, project: str, issues: bool, prs: bool) -> None:
    """
    Open one of your projects on GitHub.

    Given a project name (must exist on GitHub and be owned by you),
    'gh' will open your browser and navigate to the project on GitHub.

    You can also use the "--issues" or "--prs" flags to immediately
    open up the repo's issues or pull requests page.

    Examples:

    $ pytoil gh my_project

    $ pytoil gh my_project --issues

    $ pytoil gh my_project --prs
    """
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        exists = repo.exists_remote(api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not exists:
            printer.error(
                f"Could not find {project!r} on GitHub. Was it a typo?", exits=1
            )
        if issues:
            printer.info(f"Opening {project}'s issues on GitHub")
            click.launch(url=repo.issues_url)
        elif prs:
            printer.info(f"Opening {project}'s pull requests on GitHub")
            click.launch(url=repo.pulls_url)
        else:
            printer.info(f"Opening {project} on GitHub")
            click.launch(url=repo.html_url)



"""
The pytoil bug command.


Author: Tom Fleet
Created: 23/02/2022
"""

from __future__ import annotations

import click

from pytoil.cli.printer import printer
from pytoil.config import defaults


@click.command()
def bug() -> None:
    """
    Raise an issue about pytoil.

    The bug command let's you easily raise an issue on the pytoil
    repo. This can be a bug report, feature request, or a question!

    Examples:

    $ pytoil bug
    """
    printer.info("Opening pytoil's issues in your browser...")
    click.launch(url=defaults.PYTOIL_ISSUES_URL)



from __future__ import annotations

import os
import platform
import tempfile
from pathlib import Path

import pytest

from pytoil.config import Config, defaults

# GitHub Actions
ON_CI = bool(os.getenv("CI"))
ON_WINDOWS = platform.system().lower() == "windows"


def test_config_init_defaults():
    config = Config()

    assert config.projects_dir == defaults.PROJECTS_DIR
    assert config.token == defaults.TOKEN
    assert config.username == defaults.USERNAME
    assert config.editor == defaults.EDITOR
    assert config.conda_bin == defaults.CONDA_BIN
    assert config.common_packages == defaults.COMMON_PACKAGES
    assert config.git == defaults.GIT


def test_config_init_passed():
    config = Config(
        projects_dir=Path("some/dir"),
        token="sometoken",
        username="me",
        editor="fakeedit",
        conda_bin="mamba",
        common_packages=["black", "mypy", "flake8"],
        git=False,
    )

    assert config.projects_dir == Path("some/dir")
    assert config.token == "sometoken"
    assert config.username == "me"
    assert config.editor == "fakeedit"
    assert config.conda_bin == "mamba"
    assert config.common_packages == ["black", "mypy", "flake8"]
    assert config.git is False


def test_config_helper():
    config = Config.helper()

    assert config.projects_dir == defaults.PROJECTS_DIR
    assert config.token == "Put your GitHub personal access token here"
    assert config.username == "This your GitHub username"
    assert config.editor == defaults.EDITOR
    assert config.conda_bin == defaults.CONDA_BIN
    assert config.common_packages == defaults.COMMON_PACKAGES
    assert config.git == defaults.GIT


def test_config_load():
    with tempfile.NamedTemporaryFile(
        "w", delete=False if ON_CI and ON_WINDOWS else True
    ) as file:
        # Make a fake config object
        config = Config(
            projects_dir=Path("~/some/dir"),
            token="sometoken",
            username="me",
            editor="fakeedit",
            common_packages=["black", "mypy", "flake8"],
            git=False,
        )

        # Write the config
        config.write(path=Path(file.name))

        # Load the config
        loaded_config = Config.load(path=Path(file.name))

    assert loaded_config.projects_dir == Path("~/some/dir").expanduser()
    assert loaded_config.token == "sometoken"
    assert loaded_config.username == "me"
    assert loaded_config.editor == "fakeedit"
    assert loaded_config.common_packages == ["black", "mypy", "flake8"]
    assert loaded_config.git is False


@pytest.mark.parametrize(
    "editor, want",
    [
        ("code", True),
        ("", True),  # Because it will default to $EDITOR
        ("None", False),
        ("none", False),
    ],
)
def test_specifies_editor(editor: str, want: bool):
    config = Config(editor=editor)
    assert config.specifies_editor() is want


def test_from_file_raises_on_missing_file():
    with pytest.raises(FileNotFoundError):
        Config.load(path=Path("not/here.toml"))


@pytest.mark.parametrize(
    "username, token, expected",
    [
        ("", "", False),
        ("", "something", False),
        ("something", "", False),
        (
            "This your GitHub username",
            "Put your GitHub personal access token here",
            False,
        ),
        ("", "Put your GitHub personal access token here", False),
        ("This your GitHub username", "", False),
        ("something", "something", True),
    ],
)
def test_can_use_api(username: str, token: str, expected: bool):
    config = Config(username=username, token=token)

    assert config.can_use_api() is expected


def test_file_write():
    with tempfile.NamedTemporaryFile(
        "w", delete=False if ON_CI and ON_WINDOWS else True
    ) as file:
        # Make a fake config object
        config = Config(
            projects_dir=Path("some/dir").expanduser().resolve(),
            token="sometoken",
            username="me",
            editor="fakeedit",
            common_packages=["black", "mypy", "flake8"],
            git=False,
        )

        # Write the config
        config.write(path=Path(file.name))

        file_config = Config.load(Path(file.name))

        assert file_config == config



from __future__ import annotations

from click.testing import CliRunner

from pytoil.cli.root import main


def test_cli_doesnt_blow_up():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])

    assert result.exit_code == 0
    assert "Helpful CLI to automate the development workflow" in result.stdout



from __future__ import annotations

import sys
import tempfile
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.exceptions import GoNotInstalledError
from pytoil.starters import GoStarter


def test_go_starter_init():
    starter = GoStarter(path=Path("somewhere"), name="testygo", go="notgo")

    assert starter.path == Path("somewhere")
    assert starter.name == "testygo"
    assert starter.root == Path("somewhere").joinpath("testygo").resolve()
    assert starter.files == [
        Path("somewhere").joinpath("testygo").resolve().joinpath("README.md"),
        Path("somewhere").joinpath("testygo").resolve().joinpath("main.go"),
    ]


def test_generate_raises_if_go_not_installed():
    starter = GoStarter(path=Path("somewhere"), name="testygo", go=None)

    with pytest.raises(GoNotInstalledError):
        starter.generate()


def test_go_starter_generate(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmpdir:
        starter = GoStarter(path=Path(tmpdir), name="tempgo", go="notgo")

        mock_go_mod_init = mocker.patch(
            "pytoil.starters.go.subprocess.run", autospec=True
        )

        starter.generate(username="me")

        mock_go_mod_init.assert_called_once_with(
            ["notgo", "mod", "init", "github.com/me/tempgo"],
            cwd=starter.root,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )

        for file in starter.files:
            assert file.exists()

        with open(starter.root.joinpath("README.md")) as readme:
            readme_content = readme.read()

        with open(starter.root.joinpath("main.go")) as main_go:
            main_go_content = main_go.read()

        assert readme_content == "# tempgo\n"
        assert (
            main_go_content
            == 'package main\n\nimport "fmt"\n\nfunc main() {\n\tfmt.Println("Hello'
            ' World")\n}\n'
        )



from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path
from typing import NamedTuple

import pytest
from pytest_mock import MockerFixture

from pytoil.environments import Conda
from pytoil.exceptions import (
    BadEnvironmentFileError,
    CondaNotInstalledError,
    EnvironmentAlreadyExistsError,
    EnvironmentDoesNotExistError,
    UnsupportedCondaInstallationError,
)


def test_conda_default(mocker: MockerFixture):
    conda = Conda(root=Path("somewhere"), environment_name="test")

    # Mock the return of get_envs_dir so it thinks it exists regardless
    # of whether the tester has conda installed or not
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("anaconda3/envs"),
    )

    assert conda.project_path == Path("somewhere").resolve()
    assert conda.executable == Path("anaconda3/envs").joinpath("test/bin/python")
    assert conda.name == "conda"
    assert conda.environment_name == "test"
    assert conda.conda == shutil.which("conda")


def test_conda_passed(mocker: MockerFixture):
    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    # Mock the return of get_envs_dir so it thinks it exists regardless
    # of whether the tester has conda installed or not
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("anaconda3/envs"),
    )

    assert conda.project_path == Path("somewhere").resolve()
    assert conda.executable == Path("anaconda3/envs").joinpath("test/bin/python")
    assert conda.name == "conda"
    assert conda.environment_name == "test"
    assert conda.conda == "notconda"


def test_conda_repr():
    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")
    assert (
        repr(conda)
        == f"Conda(root={Path('somewhere')!r}, environment_name='test',"
        " conda='notconda')"
    )


@pytest.mark.parametrize(
    "exists_return, want",
    [
        (True, True),
        (False, False),
    ],
)
def test_exists(mocker: MockerFixture, exists_return: bool, want: bool):
    mocker.patch(
        "pytoil.environments.conda.Path.exists",
        autospec=True,
        return_value=exists_return,
    )

    # Mock the return of get_envs_dir so it thinks it exists regardless
    # of whether the tester has conda installed or not
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("anaconda3/envs"),
    )

    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    assert conda.exists() is want


def test_create_raises_if_conda_not_installed():
    conda = Conda(root=Path("somewhere"), environment_name="test", conda=None)

    with pytest.raises(CondaNotInstalledError):
        conda.create()


def test_create_raises_if_environment_already_exists(mocker: MockerFixture):
    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    # Make it think our "test" environment already exists
    mocker.patch(
        "pytoil.environments.conda.Conda.exists", autospec=True, return_value=True
    )

    with pytest.raises(EnvironmentAlreadyExistsError):
        conda.create()


@pytest.mark.parametrize(
    "packages, silent, stdout, stderr",
    [
        (
            ["black", "mypy", "isort"],
            True,
            subprocess.DEVNULL,
            subprocess.DEVNULL,
        ),
        (
            ["black", "mypy", "isort"],
            False,
            sys.stdout,
            sys.stderr,
        ),
    ],
)
def test_create_correctly_adds_packages_if_specified(
    mocker: MockerFixture, packages: list[str], silent: bool, stdout, stderr
):
    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    # Mock the return of get_envs_dir so it thinks it exists regardless
    # of whether the tester has conda installed or not
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("anaconda3/envs"),
    )

    conda.create(packages=packages, silent=silent)

    mock_subprocess.assert_called_once_with(
        ["notconda", "create", "-y", "--name", "test", "python=3", *packages],
        cwd=conda.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (
            True,
            subprocess.DEVNULL,
            subprocess.DEVNULL,
        ),
        (
            False,
            sys.stdout,
            sys.stderr,
        ),
    ],
)
def test_create_doesnt_add_packages_if_not_specified(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    # Mock the return of get_envs_dir so it thinks it exists regardless
    # of whether the tester has conda installed or not
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("anaconda3/envs"),
    )

    conda.create(silent=silent)

    mock_subprocess.assert_called_once_with(
        ["notconda", "create", "-y", "--name", "test", "python=3"],
        cwd=conda.project_path,
        stdout=stdout,
        stderr=stderr,
    )


def test_create_from_yml_raises_if_conda_not_installed(mocker: MockerFixture):
    conda = Conda(root=Path("somewhere"), environment_name="test")

    # Ensure shutil.which returns None
    mocker.patch(
        "pytoil.environments.conda.shutil.which",
        autospec=True,
        return_value=None,
    )

    with pytest.raises(CondaNotInstalledError):
        conda.create_from_yml(Path("somewhere"), conda="notconda")


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_create_from_yml_correctly_calls_subprocess(
    mocker: MockerFixture, temp_environment_yml: Path, silent: bool, stdout, stderr
):
    # Mock out the actual call to conda
    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    # Give it a fake envs dir
    mocker.patch(
        "pytoil.environments.conda.Conda.get_envs_dir",
        autospec=True,
        return_value=Path("/Users/testyconda3/envs"),
    )

    # Ensure shutil.which doesn't fail
    mocker.patch(
        "pytoil.environments.conda.shutil.which",
        autospec=True,
        return_value="conda",
    )

    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    conda.create_from_yml(
        project_path=temp_environment_yml.parent, conda="notconda", silent=silent
    )

    mock_subprocess.assert_called_once_with(
        ["notconda", "env", "create", "--file", f"{temp_environment_yml.resolve()}"],
        cwd=temp_environment_yml.resolve().parent,
        stdout=stdout,
        stderr=stderr,
    )


def test_create_from_yml_raises_on_bad_yml_file(
    mocker: MockerFixture, bad_temp_environment_yml: Path
):
    # Ensure shutil.which doesn't fail
    mocker.patch(
        "pytoil.environments.conda.shutil.which",
        autospec=True,
        return_value="notconda",
    )

    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    with pytest.raises(BadEnvironmentFileError):
        conda.create_from_yml(bad_temp_environment_yml.parent, conda="notconda")


def test_create_from_yml_raises_if_environment_already_exists(
    mocker: MockerFixture, temp_environment_yml: Path
):
    # Ensure shutil.which doesn't fail
    mocker.patch(
        "pytoil.environments.conda.shutil.which",
        autospec=True,
        return_value="notconda",
    )

    # Make it think the environment already exists
    mocker.patch(
        "pytoil.environments.conda.Conda.exists", autospec=True, return_value=True
    )

    conda = Conda(root=Path("somewhere"), environment_name="test", conda="notconda")

    with pytest.raises(EnvironmentAlreadyExistsError):
        conda.create_from_yml(
            project_path=temp_environment_yml.parent, conda="notconda"
        )


def test_get_envs_dir_returns_correctly_for_miniconda(
    mocker: MockerFixture, fake_home_folder_miniconda: Path
):
    mocker.patch(
        "pytoil.environments.conda.Path.home",
        autospec=True,
        return_value=fake_home_folder_miniconda,
    )

    env = Conda.get_envs_dir()

    expected_env_dir = fake_home_folder_miniconda.joinpath("miniconda3/envs")

    assert env == expected_env_dir


def test_get_envs_dir_returns_correctly_for_anaconda(
    mocker: MockerFixture, fake_home_folder_anaconda: Path
):
    mocker.patch(
        "pytoil.environments.conda.Path.home",
        autospec=True,
        return_value=fake_home_folder_anaconda,
    )

    env = Conda.get_envs_dir()

    expected_env_dir = fake_home_folder_anaconda.joinpath("anaconda3/envs")

    assert env == expected_env_dir


def test_get_envs_dir_returns_correctly_for_miniforge(
    mocker: MockerFixture, fake_home_folder_miniforge: Path
):
    mocker.patch(
        "pytoil.environments.conda.Path.home",
        autospec=True,
        return_value=fake_home_folder_miniforge,
    )

    env = Conda.get_envs_dir()

    expected_env_dir = fake_home_folder_miniforge.joinpath("miniforge3/envs")

    assert env == expected_env_dir


def test_get_envs_dir_returns_correctly_for_mambaforge(
    mocker: MockerFixture, fake_home_folder_mambaforge: Path
):
    mocker.patch(
        "pytoil.environments.conda.Path.home",
        autospec=True,
        return_value=fake_home_folder_mambaforge,
    )

    env = Conda.get_envs_dir()

    expected_env_dir = fake_home_folder_mambaforge.joinpath("mambaforge/envs")

    assert env == expected_env_dir


def test_get_envs_dir_raises_if_none_found(
    mocker: MockerFixture, fake_home_folder_no_conda: Path
):
    mocker.patch(
        "pytoil.environments.conda.Path.home",
        autospec=True,
        return_value=fake_home_folder_no_conda,
    )

    with pytest.raises(UnsupportedCondaInstallationError):
        Conda.get_envs_dir()


@pytest.mark.parametrize(
    "name, packages, silent, stdout, stderr",
    [
        (
            "sillyenv",
            ["numpy", "pandas", "requests"],
            True,
            subprocess.DEVNULL,
            subprocess.DEVNULL,
        ),
        (
            "sillyenv",
            ["numpy", "pandas", "requests"],
            False,
            sys.stdout,
            sys.stderr,
        ),
    ],
)
def test_install_passes_correct_command(
    mocker: MockerFixture, name: str, packages: list[str], silent: bool, stdout, stderr
):
    fake_project = Path("/Users/me/projects/fakeproject")

    mocker.patch("pytoil.environments.Conda.exists", autospec=True, return_value=True)

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run", autospec=True
    )

    env = Conda(root=fake_project, environment_name="testy", conda="notconda")

    env.install(packages=packages, silent=silent)

    mock_subprocess.assert_called_once_with(
        ["notconda", "install", "-y", "--name", "testy", *packages],
        cwd=fake_project.resolve(),
        stdout=stdout,
        stderr=stderr,
    )


def test_install_raises_if_environment_doesnt_exist(mocker: MockerFixture):
    mocker.patch("pytoil.environments.Conda.exists", autospec=True, return_value=False)

    conda = Conda(root=Path("somewhere"), environment_name="testy", conda="notconda")

    with pytest.raises(EnvironmentDoesNotExistError):
        conda.install(packages=["black", "mypy", "isort"])


def test_install_raises_if_conda_not_installed():
    conda = Conda(root=Path("somewhere"), environment_name="testy", conda=None)

    with pytest.raises(CondaNotInstalledError):
        conda.install(packages=["black", "mypy", "isort"])


@pytest.mark.parametrize("silent", [True, False])
def test_install_self_calls_create_from_yml(mocker: MockerFixture, silent: bool):
    mock_create_from_yml = mocker.patch(
        "pytoil.environments.conda.Conda.create_from_yml", autospec=True
    )

    conda = Conda(root=Path("somewhere"), environment_name="testy", conda="notconda")

    conda.install_self(silent=silent)

    mock_create_from_yml.assert_called_once_with(
        project_path=Path("somewhere").resolve(), silent=silent, conda="notconda"
    )


@pytest.mark.parametrize("silent", [True, False])
def test_install_self_raises_if_conda_not_installed(
    mocker: MockerFixture, silent: bool
):
    mocker.patch("pytoil.environments.conda.Conda.create_from_yml", autospec=True)

    conda = Conda(root=Path("somewhere"), environment_name="testy", conda=None)

    with pytest.raises(CondaNotInstalledError):
        conda.install_self(silent=silent)


def test_export_yml_raises_if_conda_not_installed():
    conda = Conda(root=Path("somewhere"), environment_name="testy", conda=None)

    with pytest.raises(CondaNotInstalledError):
        conda.export_yml()


def test_export_yml_raises_on_missing_env(mocker: MockerFixture):
    mocker.patch(
        "pytoil.environments.conda.Conda.exists",
        autospec=True,
        return_value=False,
    )

    with pytest.raises(EnvironmentDoesNotExistError):
        env = Conda(root=Path("somewhere"), environment_name="testy", conda="notconda")
        env.export_yml()


def test_export_yml(mocker: MockerFixture, temp_environment_yml: Path):
    # It must think the environment exists
    mocker.patch(
        "pytoil.environments.conda.Conda.exists",
        autospec=True,
        return_value=True,
    )

    conda = Conda(
        root=temp_environment_yml.parent, environment_name="testy", conda="notconda"
    )

    class Process(NamedTuple):
        content: str = """
        name: testy
        channels:
        - defaults
        - conda-forge
        dependencies:
        - python=3
        - invoke
        - black
        - flake8
        - isort
        - mypy
        - rich
        - numpy
        - requests
        - pandas
        """

        @property
        def stdout(self) -> str:
            return self.content

    mock_subprocess = mocker.patch(
        "pytoil.environments.conda.subprocess.run",
        autospec=True,
        return_value=Process(),
    )

    conda.export_yml()

    mock_subprocess.assert_called_once_with(
        ["notconda", "env", "export", "--from-history", "--name", "testy"],
        cwd=temp_environment_yml.parent.resolve(),
        capture_output=True,
        encoding="utf-8",
    )

    assert temp_environment_yml.read_text(encoding="utf-8") == Process().content



"""
The root CLI command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

from pathlib import Path

import click
import questionary
import rich.traceback

from pytoil import __version__
from pytoil.cli.bug import bug
from pytoil.cli.checkout import checkout
from pytoil.cli.config import config
from pytoil.cli.docs import docs
from pytoil.cli.find import find
from pytoil.cli.gh import gh
from pytoil.cli.info import info
from pytoil.cli.keep import keep
from pytoil.cli.new import new
from pytoil.cli.printer import printer
from pytoil.cli.pull import pull
from pytoil.cli.remove import remove
from pytoil.cli.show import show
from pytoil.config import Config, defaults

# So that if we do ever get a traceback, it uses rich to show it nicely
rich.traceback.install()


@click.group(
    commands={
        "checkout": checkout,
        "config": config,
        "docs": docs,
        "find": find,
        "gh": gh,
        "info": info,
        "new": new,
        "pull": pull,
        "remove": remove,
        "show": show,
        "keep": keep,
        "bug": bug,
    }
)
@click.version_option(version=__version__, prog_name="pytoil")
@click.pass_context
def main(ctx: click.Context) -> None:
    """
    Helpful CLI to automate the development workflow.

    - Create and manage your local and remote projects

    - Build projects from cookiecutter templates.

    - Easily create/manage virtual environments.

    - Minimal configuration required.
    """
    # Load the config once on launch of the app and pass it down to the child commands
    # through click's context
    try:
        config = Config.load()
    except FileNotFoundError:
        interactive_config()
    else:
        if not config.can_use_api():
            printer.error(
                (
                    "You must set your GitHub username and personal access token to use"
                    " API features."
                ),
                exits=1,
            )
        # We have a valid config file at the right place so load it into click's
        # context and pass it down to all subcommands
        ctx.obj = config


def interactive_config() -> None:
    """
    Prompt the user with a series of questions
    to configure pytoil interactively.
    """
    printer.warn("No pytoil config file detected!")
    interactive: bool = questionary.confirm(
        "Interactively configure pytoil?", default=False, auto_enter=False
    ).ask()

    if not interactive:
        # User doesn't want to interactively walk through a config file
        # so just make a default and exit cleanly
        Config.helper().write()
        printer.good("I made a default file for you.")
        printer.note(
            (
                f"It's here: {defaults.CONFIG_FILE}, you can edit it with `pytoil"
                " config edit``"
            ),
            exits=0,
        )
        return

    # If we get here, the user wants to interactively make the config
    projects_dir: str = questionary.path(
        "Where do you keep your projects?",
        default=str(defaults.PROJECTS_DIR),
        only_directories=True,
    ).ask()

    token: str = questionary.text("GitHub personal access token?").ask()

    username: str = questionary.text("What's your GitHub username?").ask()

    use_editor: bool = questionary.confirm(
        "Auto open projects in an editor?", default=False, auto_enter=False
    ).ask()

    if use_editor:
        editor: str = questionary.text("Name of the editor binary to use?").ask()
    else:
        editor = "None"

    git: bool = questionary.confirm(
        "Make git repos when creating new projects?", default=True, auto_enter=False
    ).ask()

    conda_bin: str = questionary.select(
        "Use conda or mamba for conda environments?",
        choices=("conda", "mamba"),
        default="conda",
    ).ask()

    config = Config(
        projects_dir=Path(projects_dir).resolve(),
        token=token,
        username=username,
        editor=editor,
        conda_bin=conda_bin,
        git=git,
    )

    config.write()

    printer.good("Config created")
    printer.note(f"It's available at {defaults.CONFIG_FILE}.", exits=0)



from __future__ import annotations

import tempfile
from pathlib import Path

from pytoil.starters import PythonStarter


def test_python_starter_init():
    starter = PythonStarter(path=Path("somewhere"), name="testypython")

    assert starter.path == Path("somewhere")
    assert starter.name == "testypython"
    assert starter.root == Path("somewhere").joinpath("testypython").resolve()
    assert starter.files == [
        Path("somewhere").joinpath("testypython").resolve().joinpath("README.md"),
        Path("somewhere")
        .joinpath("testypython")
        .resolve()
        .joinpath("requirements.txt"),
        Path("somewhere").joinpath("testypython").resolve().joinpath("testypython.py"),
    ]


def test_python_starter_generate():
    with tempfile.TemporaryDirectory() as tmpdir:
        starter = PythonStarter(path=Path(tmpdir), name="temptest")

        starter.generate()

        for file in starter.files:
            assert file.exists()

        with open(starter.root.joinpath("README.md")) as readme:
            readme_content = readme.read()

        with open(starter.root.joinpath("requirements.txt")) as requirements:
            requirements_content = requirements.read()

        with open(starter.root.joinpath("temptest.py")) as python:
            python_content = python.read()

        assert readme_content == "# temptest\n"
        assert (
            requirements_content == "# Put your requirements here e.g. flask>=1.0.0\n"
        )
        assert (
            python_content
            == 'def hello(name: str = "world") -> None:\n    print(f"hello {name}")\n'
        )



from __future__ import annotations

from pathlib import Path

import pytest
from freezegun import freeze_time
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from pytoil.api import API
from pytoil.config import Config
from pytoil.environments import Conda, Flit, Poetry, Requirements, Venv
from pytoil.repo import Repo

PROJECT_ROOT = Path(__file__).parent.parent.resolve()


def test_clone_url():
    repo = Repo(owner="me", name="project", local_path=Path("doesn't/matter"))

    assert repo.clone_url == "https://github.com/me/project.git"


def test_html_url():
    repo = Repo(owner="me", name="project", local_path=Path("doesn't/matter"))

    assert repo.html_url == "https://github.com/me/project"


def test_issues_url():
    repo = Repo(owner="me", name="project", local_path=Path("doesn't/matter"))

    assert repo.issues_url == "https://github.com/me/project/issues"


def test_pulls_url():
    repo = Repo(owner="me", name="project", local_path=Path("doesn't/matter"))

    assert repo.pulls_url == "https://github.com/me/project/pulls"


def test_repo_repr():
    repo = Repo(owner="me", name="project", local_path=Path("somewhere"))
    assert (
        repr(repo)
        == f"Repo(owner='me', name='project', local_path={Path('somewhere')!r})"
    )


@pytest.mark.parametrize(
    "path, exists",
    [
        (PROJECT_ROOT, True),  # This project should hopefully exist!
        (Path("not/here"), False),
    ],
)
def test_exists_local(path: Path, exists: bool):
    repo = Repo(owner="me", name="test", local_path=path)

    assert repo.exists_local() is exists


def test_exists_remote_returns_false_when_remote_doesnt_exist(
    httpx_mock: HTTPXMock, fake_repo_exists_false_response
):
    api = API(username="me", token="something")

    repo = Repo(owner="me", name="test", local_path=Path("doesn't/matter"))

    httpx_mock.add_response(
        url=api.url, json=fake_repo_exists_false_response, status_code=200
    )

    result = repo.exists_remote(api=api)
    assert result is False


def test_exists_remote_returns_true_when_remote_exists(
    httpx_mock: HTTPXMock, fake_repo_exists_true_response
):
    api = API(username="me", token="something")

    repo = Repo(owner="me", name="test", local_path=Path("doesn't/matter"))

    httpx_mock.add_response(
        url=api.url, json=fake_repo_exists_true_response, status_code=200
    )

    result = repo.exists_remote(api=api)
    assert result is True


@freeze_time("2022-01-22 09:00")
def test_remote_info_returns_correct_details(
    httpx_mock: HTTPXMock, fake_repo_info_response
):
    api = API(username="me", token="something")
    repo = Repo(owner="me", name="test", local_path=Path("somewhere"))

    httpx_mock.add_response(url=api.url, json=fake_repo_info_response, status_code=200)

    result = repo._remote_info(api)
    assert result == {
        "Created": "11 months ago",
        "Description": "CLI to automate the development workflow :robot:",
        "License": "Apache License 2.0",
        "Name": "pytoil",
        "Remote": True,
        "Size": "3.2 MB",
        "Updated": "25 days ago",
        "Language": "Python",
    }


@pytest.mark.parametrize(
    "file, exists",
    [
        ("here.txt", True),
        ("i_exist.yml", True),
        ("hello.py", True),
        ("me_too.json", True),
        ("not_me_though.csv", False),
        ("me_neither.toml", False),
        ("i_never_existed.cfg", False),
        ("what_file.ini", False),
    ],
)
def test_does_file_exist(
    repo_folder_with_random_existing_files: Path,
    file: str,
    exists: bool,
):
    repo = Repo(
        owner="me", name="test", local_path=repo_folder_with_random_existing_files
    )

    assert repo._file_exists(file) is exists


@pytest.mark.parametrize(
    "file, expect",
    [
        ("setup.cfg", True),
        ("setup.py", True),
        ("dingle.cfg", False),
        ("dingle.py", False),
        ("pyproject.toml", False),  # Empty pyproject.toml
        ("environment.yml", False),
    ],
)
def test_is_setuptools(
    repo_folder_with_random_existing_files: Path, file: str, expect: bool
):
    repo = Repo(
        owner="me", name="test", local_path=repo_folder_with_random_existing_files
    )

    # Add in the required file to trigger
    repo_folder_with_random_existing_files.joinpath(file).touch()

    assert repo.is_setuptools() is expect


def test_is_setuptools_pep621(project_with_setuptools_pep621_backend: Path):
    repo = Repo(
        owner="me", name="test", local_path=project_with_setuptools_pep621_backend
    )

    assert repo.is_setuptools() is True


@pytest.mark.parametrize(
    "file, expect",
    [
        ("setup.cfg", False),
        ("setup.py", False),
        ("dingle.cfg", False),
        ("dingle.py", False),
        ("pyproject.toml", False),
        ("environment.yml", True),
    ],
)
def test_is_conda(
    repo_folder_with_random_existing_files: Path, file: str, expect: bool
):
    repo = Repo(
        owner="me", name="test", local_path=repo_folder_with_random_existing_files
    )
    # Add in the required file to trigger
    repo_folder_with_random_existing_files.joinpath(file).touch()

    assert repo.is_conda() is expect


def test_is_poetry_true_on_valid_poetry_project(fake_poetry_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_poetry_project)

    assert repo.is_poetry() is True


def test_is_poetry_false_on_non_poetry_project(fake_flit_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_flit_project)

    assert repo.is_poetry() is False


def test_is_poetry_false_if_no_pyproject_toml():
    repo = Repo(owner="blah", name="test", local_path=Path("nowhere"))

    assert repo.is_poetry() is False


def test_is_poetry_false_if_no_build_system(project_with_no_build_system: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_system)

    assert repo.is_poetry() is False


def test_is_poetry_false_if_no_build_backend(project_with_no_build_backend: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_backend)

    assert repo.is_poetry() is False


def test_is_flit_true_on_valid_flit_project(fake_flit_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_flit_project)

    assert repo.is_flit() is True


def test_is_flit_false_on_non_flit_project(fake_poetry_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_poetry_project)

    assert repo.is_flit() is False


def test_is_flit_false_if_no_pyproject_toml():
    repo = Repo(owner="blah", name="test", local_path=Path("nowhere"))

    assert repo.is_flit() is False


def test_is_flit_false_if_no_build_system(project_with_no_build_system: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_system)

    assert repo.is_flit() is False


def test_is_flit_false_if_no_build_backend(project_with_no_build_backend: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_backend)

    assert repo.is_flit() is False


def test_is_hatch_true_on_valid_hatch_project(fake_hatch_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_hatch_project)

    assert repo.is_hatch() is True


def test_is_hatch_false_on_non_hatch_project(fake_poetry_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_poetry_project)

    assert repo.is_hatch() is False


def test_is_hatch_false_if_no_pyproject_toml():
    repo = Repo(owner="blah", name="test", local_path=Path("nowhere"))

    assert repo.is_hatch() is False


def test_is_hatch_false_if_no_build_system(project_with_no_build_system: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_system)

    assert repo.is_hatch() is False


def test_is_hatch_false_if_no_build_backend(project_with_no_build_backend: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_backend)

    assert repo.is_hatch() is False


def test_is_pep621_true_on_valid_pep621_project(fake_pep621_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_pep621_project)

    assert repo.is_pep621() is True


def test_is_pep621_false_on_non_pep621_project(fake_poetry_project: Path):
    repo = Repo(owner="blah", name="test", local_path=fake_poetry_project)

    assert repo.is_pep621() is False


def test_is_pep621_false_if_no_pyproject_toml():
    repo = Repo(owner="blah", name="test", local_path=Path("nowhere"))

    assert repo.is_pep621() is False


def test_is_pep621_false_if_no_build_system(project_with_no_build_system: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_system)

    assert repo.is_pep621() is False


def test_is_pep621_false_if_no_build_backend(project_with_no_build_backend: Path):
    repo = Repo(owner="blah", name="test", local_path=project_with_no_build_backend)

    assert repo.is_pep621() is False


def test_dispatch_env_correctly_identifies_conda(mocker: MockerFixture):
    mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=True)

    repo = Repo(name="test", owner="me", local_path=Path("somewhere"))

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Conda)


def test_dispatch_env_correctly_identifies_requirements_txt(
    requirements_project: Path,
):
    repo = Repo(name="test", owner="me", local_path=requirements_project)

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Requirements)


def test_dispatch_env_correctly_identifies_requirements_dev_txt(
    requirements_dev_project: Path,
):
    repo = Repo(name="test", owner="me", local_path=requirements_dev_project)

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Requirements)


def test_dispatch_env_correctly_identifies_setuptools(mocker: MockerFixture):
    mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=False)

    mocker.patch("pytoil.repo.Repo.is_setuptools", autospec=True, return_value=True)

    repo = Repo(name="test", owner="me", local_path=Path("somewhere"))

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Venv)


def test_dispatch_env_correctly_identifies_poetry(fake_poetry_project: Path):
    repo = Repo(name="test", owner="me", local_path=fake_poetry_project)

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Poetry)


def test_dispatch_env_correctly_identifies_flit(fake_flit_project: Path):
    repo = Repo(name="test", owner="me", local_path=fake_flit_project)

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Flit)


def test_dispatch_env_correctly_identifies_pep621(fake_pep621_project: Path):
    repo = Repo(name="test", owner="me", local_path=fake_pep621_project)

    env = repo.dispatch_env(config=Config())

    assert isinstance(env, Venv)


def test_dispatch_env_returns_none_if_it_cant_detect(mocker: MockerFixture):
    mocker.patch("pytoil.repo.Repo.is_conda", autospec=True, return_value=False)

    mocker.patch("pytoil.repo.Repo.is_setuptools", autospec=True, return_value=False)

    repo = Repo(name="test", owner="me", local_path=Path("somewhere"))

    env = repo.dispatch_env(config=Config())

    assert env is None



from __future__ import annotations

from freezegun import freeze_time
from pytest_httpx import HTTPXMock

from pytoil import __version__
from pytoil.api import API


def test_headers():
    api = API(username="me", token="notatoken")

    assert api.headers == {
        "Authorization": "token notatoken",
        "User-Agent": f"pytoil/{__version__}",
        "Accept": "application/vnd.github.v4+json",
    }


def test_get_repo_names(httpx_mock: HTTPXMock, fake_get_repo_names_response):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(
        url=api.url, json=fake_get_repo_names_response, status_code=200
    )

    names = api.get_repo_names()

    assert names == {
        "dingle",
        "dangle",
        "dongle",
        "a_cool_project",
        "another",
        "yetanother",
        "hello",
    }


def test_check_repo_exists_returns_false_if_not_exists(
    httpx_mock: HTTPXMock, fake_repo_exists_false_response
):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(
        url=api.url, json=fake_repo_exists_false_response, status_code=200
    )

    exists = api.check_repo_exists(owner="me", name="dave")

    assert exists is False


def test_check_repo_exists_returns_true_if_exists(
    httpx_mock: HTTPXMock, fake_repo_exists_true_response
):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(
        url=api.url, json=fake_repo_exists_true_response, status_code=200
    )

    exists = api.check_repo_exists(owner="me", name="pytoil")

    assert exists is True


@freeze_time("2022-01-16")
def test_get_repo_info_good_response(httpx_mock: HTTPXMock, fake_repo_info_response):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(url=api.url, json=fake_repo_info_response, status_code=200)

    info = api.get_repo_info(name="pytoil")

    assert info == {
        "Name": "pytoil",
        "Description": "CLI to automate the development workflow :robot:",
        "Created": "11 months ago",
        "Updated": "19 days ago",
        "Size": "3.2 MB",
        "License": "Apache License 2.0",
        "Remote": True,
        "Language": "Python",
    }


@freeze_time("2022-01-16")
def test_get_repo_info_no_license(
    httpx_mock: HTTPXMock, fake_repo_info_response_no_license
):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(
        url=api.url, json=fake_repo_info_response_no_license, status_code=200
    )

    info = api.get_repo_info(name="pytoil")

    assert info == {
        "Name": "pytoil",
        "Description": "CLI to automate the development workflow :robot:",
        "Created": "11 months ago",
        "Updated": "19 days ago",
        "Size": "3.2 MB",
        "License": None,
        "Remote": True,
        "Language": "Python",
    }


def test_create_fork(httpx_mock: HTTPXMock):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(
        url="https://api.github.com/repos/someoneelse/project/forks", status_code=201
    )

    api.create_fork(owner="someoneelse", repo="project")


def test_get_repos(httpx_mock: HTTPXMock, fake_get_repos_response):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(url=api.url, json=fake_get_repos_response, status_code=200)

    data = api.get_repos()

    assert data == [
        {
            "name": "advent_of_code_2020",
            "description": "Retroactively doing AOC2020 in Go.",
            "createdAt": "2022-01-05T16:54:03Z",
            "pushedAt": "2022-01-09T06:55:32Z",
            "diskUsage": 45,
        },
        {
            "name": "advent_of_code_2021",
            "description": "My code for AOC 2021",
            "createdAt": "2021-11-30T12:01:22Z",
            "pushedAt": "2021-12-19T15:10:07Z",
            "diskUsage": 151,
        },
        {
            "name": "aircraft_crashes",
            "description": "Analysis of aircraft crash data.",
            "createdAt": "2021-01-02T19:34:15Z",
            "pushedAt": "2021-01-20T10:35:57Z",
            "diskUsage": 2062,
        },
        {
            "name": "cookie_pypackage",
            "description": "My own version of the Cookiecutter pypackage template",
            "createdAt": "2020-07-04T10:05:36Z",
            "pushedAt": "2021-12-03T08:45:49Z",
            "diskUsage": 734,
        },
        {
            "name": "cv",
            "description": "Repo for my CV, built with JSON Resume.",
            "createdAt": "2021-10-30T15:11:49Z",
            "pushedAt": "2022-01-10T19:50:24Z",
            "diskUsage": 145,
        },
        {
            "name": "eu_energy_analysis",
            "description": "Analysis of the EU Open Power System Data.",
            "createdAt": "2020-12-13T10:50:35Z",
            "pushedAt": "2020-12-24T11:12:34Z",
            "diskUsage": 1834,
        },
        {
            "name": "FollowTheProcess",
            "description": 'My "About Me" Repo',
            "createdAt": "2020-07-14T16:06:52Z",
            "pushedAt": "2022-01-10T20:05:47Z",
            "diskUsage": 14640,
        },
        {
            "name": "followtheprocess.github.io",
            "description": "Repo for my GitHub pages site.",
            "createdAt": "2021-02-19T20:16:05Z",
            "pushedAt": "2021-11-18T19:04:06Z",
            "diskUsage": 10753,
        },
    ]


def test_get_forks(httpx_mock: HTTPXMock, fake_get_forks_response):
    api = API(username="me", token="definitelynotatoken")

    httpx_mock.add_response(url=api.url, json=fake_get_forks_response, status_code=200)

    data = api.get_forks()

    assert data == [
        {
            "name": "nox",
            "diskUsage": 5125,
            "createdAt": "2021-07-01T11:43:36Z",
            "pushedAt": "2022-01-08T11:00:44Z",
            "parent": {"nameWithOwner": "theacodes/nox"},
        },
        {
            "name": "python-launcher",
            "diskUsage": 824,
            "createdAt": "2021-10-25T18:33:11Z",
            "pushedAt": "2021-11-09T07:47:23Z",
            "parent": {"nameWithOwner": "brettcannon/python-launcher"},
        },
    ]



"""
The pytoil config command group.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import click
from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table

from pytoil.cli.printer import printer
from pytoil.config import Config, defaults


@click.group()
def config() -> None:
    """
    Interact with pytoil's configuration.

    The config command group allows you to get, show and explain pytoil's configuration.
    """


@config.command()
@click.pass_obj
def show(config: Config) -> None:
    """
    Show pytoil's config.

    The show command allows you to easily see pytoil's current config.

    The values are taken directly from the config file where specified or
    the defaults otherwise.

    Examples:

    $ pytoil config show
    """
    table = Table(box=box.SIMPLE)
    table.add_column("Key", style="cyan", justify="right")
    table.add_column("Value", justify="left")

    for key, val in config.to_dict().items():
        table.add_row(f"{key}:", str(val))

    console = Console()
    console.print(table)


@config.command()
@click.argument("key", nargs=1)
@click.pass_obj
def get(config: Config, key: str) -> None:
    """
    Get the currently set value for a config key.

    The get command will only allow valid pytoil config keys.

    Examples:

    $ pytoil config get editor
    """
    if key not in defaults.CONFIG_KEYS:
        printer.error(f"{key} is not a valid pytoil config key.", exits=1)

    console = Console()
    console.print(f"[cyan]{key}[/]: [default]{config.to_dict().get(key)}[/]")


@config.command()
def edit() -> None:
    """
    Open pytoil's config file in $EDITOR.

    Examples:

    $ pytoil config edit
    """
    click.launch(str(defaults.CONFIG_FILE), wait=False)


@config.command()
def explain() -> None:
    """
    Print a list and description of pytoil config values.

    Examples:

    $ pytoil config explain
    """
    console = Console()
    markdown = Markdown(defaults.CONFIG_SCHEMA, justify="center")
    console.print(markdown)



"""
Collection of useful helpers for the CLI.


Author: Tom Fleet
Created: 30/12/2021
"""
from __future__ import annotations

import httpx

from pytoil.cli.printer import printer


def handle_http_status_error(error: httpx.HTTPStatusError) -> None:
    """
    Handles a variety of possible HTTP Status errors, print's nicer output
    to the user, and exits the program if necessary.
    Call this in an except block on CLI commands accessing the
    GitHub API.
    Args:
        error (httpx.HTTPStatusError): The error to be handled.
    """
    code = error.response.status_code

    if code == 401:
        printer.error("HTTP 401 - Unauthorized")
        printer.note("This usually means something is wrong with your token!", exits=1)
    elif code == 404:
        printer.error("HTTP 404 - Not Found")
        printer.note("This is a bug we've not handled, please raise an issue!", exits=1)
    elif code == 500:
        printer.error("HTTP 500 - Server Error")
        printer.note("This is very rare but it means GitHub is not happy!", exits=1)



"""
The pytoil pull command.


Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

import click
import httpx
import questionary

from pytoil.api import API
from pytoil.cli import utils
from pytoil.cli.printer import printer
from pytoil.config import Config
from pytoil.git import Git
from pytoil.repo import Repo


@click.command()
@click.argument("projects", nargs=-1)
@click.option("-f", "--force", is_flag=True, help="Force pull without confirmation.")
@click.option("-a", "--all", "all_", is_flag=True, help="Pull down all your projects.")
@click.pass_obj
def pull(config: Config, projects: tuple[str, ...], force: bool, all_: bool) -> None:
    """
    Pull down your remote projects.

    The pull command provides easy methods for pulling down remote projects.

    It is effectively a nice wrapper around git clone but you don't have to
    worry about urls or what your cwd is, pull will grab your remote projects
    by name and clone them to your configured projects directory.

    You can also use pull to batch clone multiple repos, even all of them ("--all/-a")
    if you're into that sorta thing.

    If more than 1 repo is passed (or if "--all/-a" is used) pytoil will pull
    the repos concurrently, speeding up the process.

    Any remote project that already exists locally will be skipped and none of
    your local projects are changed in any way. pytoil will only pull down
    those projects that don't already exist locally.

    It's very possible to accidentally clone a lot of repos when using pull so
    you will be prompted for confirmation before pytoil does anything.

    The "--force/-f" flag can be used to override this confirmation prompt if
    desired.

    Examples:

    $ pytoil pull project1 project2 project3

    $ pytoil pull project1 project2 project3 --force

    $ pytoil pull --all

    $ pytoil pull --all --force
    """
    if not projects and not all_:
        printer.error(
            "If not using the '--all' flag, you must specify projects to pull.", exits=1
        )

    api = API(username=config.username, token=config.token)

    local_projects: set[str] = {
        f.name
        for f in config.projects_dir.iterdir()
        if f.is_dir() and not f.name.startswith(".")
    }

    try:
        remote_projects = api.get_repo_names()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_error(err)
    else:
        if not remote_projects:
            printer.error("You don't have any remote projects to pull.", exits=1)

        specified_remotes = remote_projects if all_ else set(projects)

        # Check for typos
        for project in projects:
            if project not in remote_projects:
                printer.error(
                    f"{project!r} not found on GitHub. Was it a typo?", exits=1
                )

        diff = specified_remotes.difference(local_projects)
        if not diff:
            printer.good("Your local and remote projects are in sync!", exits=0)

        if not force:
            if len(diff) <= 3:
                message = f"This will pull down {', '.join(diff)}. Are you sure?"
            else:
                # Too many to show nicely
                message = f"This will pull down {len(diff)} projects. Are you sure?"

            confirmed: bool = questionary.confirm(
                message, default=False, auto_enter=False
            ).ask()

            if not confirmed:
                printer.warn("Aborted", exits=1)

        # Now we're good to go
        to_clone = [
            Repo(
                owner=config.username,
                name=project,
                local_path=config.projects_dir.joinpath(project),
            )
            for project in diff
        ]
        git = Git()
        with ThreadPoolExecutor() as executor:
            for repo in to_clone:
                executor.submit(clone_and_report, repo=repo, git=git, config=config)


def clone_and_report(repo: Repo, git: Git, config: Config) -> None:
    git.clone(url=repo.clone_url, cwd=config.projects_dir)
    printer.good(f"Cloned {repo.name!r}")



"""
Module responsible for handling pytoil's programmatic
interaction with its config file.

Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

import rtoml
from pydantic import BaseModel

from pytoil.config import defaults


class Config(BaseModel):
    projects_dir: Path = defaults.PROJECTS_DIR
    token: str = defaults.TOKEN
    username: str = defaults.USERNAME
    editor: str = defaults.EDITOR
    conda_bin: str = defaults.CONDA_BIN
    common_packages: list[str] = defaults.COMMON_PACKAGES
    git: bool = defaults.GIT

    @classmethod
    def load(cls, path: Path = defaults.CONFIG_FILE) -> Config:
        """
        Reads in the ~/.pytoil.toml config file and returns
        a populated `Config` object.

        Args:
            path (Path, optional): Path to the config file.
                Defaults to defaults.CONFIG_FILE.

        Returns:
            Config: Populated `Config` object.

        Raises:
            FileNotFoundError: If config file not found.
        """
        try:
            with open(path, encoding="utf-8") as f:
                config_dict: dict[str, Any] = rtoml.load(f).get("pytoil", "")
        except FileNotFoundError:
            raise
        else:
            # This is actually covered
            if config_dict.get("projects_dir"):  # pragma: no cover
                config_dict["projects_dir"] = (
                    Path(config_dict["projects_dir"]).expanduser().resolve()
                )
            return Config(**config_dict)

    @classmethod
    def helper(cls) -> Config:
        """
        Returns a friendly placeholder object designed to be
        written to a config file as a guide to the user on what
        to fill in.

        Most of the fields will be the default but some will have
        helpful instructions.

        Returns:
            Config: Helper config object.
        """
        return Config(
            token="Put your GitHub personal access token here",
            username="This your GitHub username",
        )

    def to_dict(self) -> dict[str, Any]:
        """
        Writes out the attributes from the calling instance
        to a dictionary.
        """
        return {
            "projects_dir": str(self.projects_dir),
            "token": self.token,
            "username": self.username,
            "editor": self.editor,
            "conda_bin": self.conda_bin,
            "common_packages": self.common_packages,
            "git": self.git,
        }

    def write(self, path: Path = defaults.CONFIG_FILE) -> None:
        """
        Overwrites the config file at `path` with the attributes from
        the calling instance.

        Args:
            path (Path, optional): Config file to overwrite.
                Defaults to defaults.CONFIG_FILE.
        """
        with open(path, mode="w", encoding="utf-8") as f:
            rtoml.dump({"pytoil": self.to_dict()}, f, pretty=True)

    def can_use_api(self) -> bool:
        """
        Helper method to easily determine whether or not
        the config instance has the required elements
        to use the GitHub API.

        Returns:
            bool: True if can use API, else False.
        """
        conditions = [
            self.username == "",
            self.username == "This your GitHub username",
            self.token == "",
            self.token == "Put your GitHub personal access token here",
        ]

        return not any(conditions)

    def specifies_editor(self) -> bool:
        """
        Returns whether the user has set an editor, either directly
        or through the $EDITOR env var.

        If a user has no `editor` key in the config file, pytoil will
        use $EDITOR.

        If the key is present but is set to the literal "None", pytoil
        will not try to open projects.

        Otherwise the value of the key `editor` will be used as the name
        of the binary.

        Returns:
            bool: True if editor is not literal "None" else False.
        """
        return self.editor.lower() != "none"



from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.environments import Poetry
from pytoil.exceptions import PoetryNotInstalledError


def test_poetry_instanciation_default():
    poetry = Poetry(root=Path("somewhere"))

    assert poetry.project_path == Path("somewhere").resolve()
    assert poetry.poetry == shutil.which("poetry")
    assert poetry.name == "poetry"
    assert poetry.executable == Path("somewhere").resolve().joinpath(".venv/bin/python")


def test_poetry_instanciation_passed():
    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")

    assert poetry.project_path == Path("somewhere").resolve()
    assert poetry.poetry == "notpoetry"
    assert poetry.name == "poetry"
    assert poetry.executable == Path("somewhere").resolve().joinpath(".venv/bin/python")


def test_poetry_repr():
    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")
    assert repr(poetry) == f"Poetry(root={Path('somewhere')!r}, poetry='notpoetry')"


@pytest.mark.parametrize(
    "exists_return, exists",
    [
        (True, True),
        (False, False),
    ],
)
def test_exists_returns_correct_value(
    mocker: MockerFixture, exists_return, exists: bool
):
    # Ensure Path.exists returns what we want it to
    mocker.patch(
        "pytoil.environments.poetry.Path.exists",
        autospec=True,
        return_value=exists_return,
    )

    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")
    assert poetry.exists() is exists


def test_create_raises_not_implemented_error():
    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")

    with pytest.raises(NotImplementedError):
        poetry.create()


def test_enforce_local_config_correctly_calls_poetry(mocker: MockerFixture):
    mock = mocker.patch("pytoil.environments.poetry.subprocess.run", autospec=True)

    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")

    poetry.enforce_local_config()

    mock.assert_called_once_with(
        ["notpoetry", "config", "virtualenvs.in-project", "true", "--local"],
        cwd=poetry.project_path,
    )


def test_enforce_local_config_raises_if_poetry_not_installed():
    poetry = Poetry(root=Path("somewhere"), poetry=None)

    with pytest.raises(PoetryNotInstalledError):
        poetry.enforce_local_config()


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_correctly_calls_poetry(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mock = mocker.patch("pytoil.environments.poetry.subprocess.run", autospec=True)

    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")

    # Mock out enforce local config as `install` runs this first
    mocker.patch(
        "pytoil.environments.poetry.Poetry.enforce_local_config", autospec=True
    )

    poetry.install(packages=["black", "isort", "flake8", "mypy"], silent=silent)

    mock.assert_called_once_with(
        ["notpoetry", "add", "black", "isort", "flake8", "mypy"],
        cwd=poetry.project_path,
        stdout=stdout,
        stderr=stderr,
    )


def test_install_raises_if_poetry_not_installed():
    poetry = Poetry(root=Path("somewhere"), poetry=None)

    with pytest.raises(PoetryNotInstalledError):
        poetry.install(packages=["something", "doesn't", "matter"])


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_correctly_calls_poetry(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mock = mocker.patch("pytoil.environments.poetry.subprocess.run", autospec=True)

    poetry = Poetry(root=Path("somewhere"), poetry="notpoetry")

    # Mock out enforce local config as `install` runs this first
    mocker.patch(
        "pytoil.environments.poetry.Poetry.enforce_local_config", autospec=True
    )

    poetry.install_self(silent=silent)

    mock.assert_called_once_with(
        ["notpoetry", "install"],
        cwd=poetry.project_path,
        stdout=stdout,
        stderr=stderr,
    )


def test_install_selfraises_if_poetry_not_installed():
    poetry = Poetry(root=Path("somewhere"), poetry=None)

    with pytest.raises(PoetryNotInstalledError):
        poetry.install_self()



"""
Module responsible for handling launching a directory-aware editor
when opening a project.


Author: Tom Fleet
Created: 12/03/2022
"""

from __future__ import annotations

import subprocess
import sys
from pathlib import Path


def launch(path: Path, bin: str) -> None:
    """
    Launch a directory-aware editor from the command line binary
    `bin` to open a project with root at `path`.

    Launch assumes that the command to open a project is of the
    structure `<bin> <path>` e.g. `code ~/myproject`.

    Args:
        path (Path): Absolute path to the root of the project to open.
        bin (str): Name of the editor binary e.g. `code`.
    """
    subprocess.run([bin, path], stdout=sys.stdout, stderr=sys.stderr)



from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.environments.flit import Flit
from pytoil.exceptions import FlitNotInstalledError


def test_flit():
    flit = Flit(root=Path("somewhere"), flit="notflit")

    assert flit.project_path == Path("somewhere").resolve()
    assert flit.name == "flit"
    assert flit.executable == Path("somewhere").resolve().joinpath(".venv/bin/python")
    assert flit.flit == "notflit"


def test_flit_repr():
    flit = Flit(root=Path("somewhere"), flit="notflit")
    assert repr(flit) == f"Flit(root={Path('somewhere')!r}, flit='notflit')"


def test_raises_if_flit_not_installed():
    flit = Flit(root=Path("somewhere"), flit=None)

    with pytest.raises(FlitNotInstalledError):
        flit.install_self()


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_venv_exists(mocker: MockerFixture, silent: bool, stdout, stderr):
    mocker.patch(
        "pytoil.environments.flit.Flit.exists",
        autospec=True,
        return_value=True,
    )

    mock = mocker.patch("pytoil.environments.flit.subprocess.run", autospec=True)

    env = Flit(root=Path("somewhere"), flit="notflit")

    env.install_self(silent=silent)

    mock.assert_called_once_with(
        [
            "notflit",
            "install",
            "--deps",
            "develop",
            "--symlink",
            "--python",
            f"{env.executable}",
        ],
        cwd=env.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_venv_doesnt_exist(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mocker.patch(
        "pytoil.environments.flit.Flit.exists",
        autospec=True,
        return_value=False,
    )

    mock_create = mocker.patch("pytoil.environments.flit.Flit.create", autospec=True)

    mock = mocker.patch("pytoil.environments.flit.subprocess.run", autospec=True)

    env = Flit(root=Path("somewhere"), flit="notflit")

    env.install_self(silent=silent)

    mock_create.assert_called_once()

    mock.assert_called_once_with(
        [
            "notflit",
            "install",
            "--deps",
            "develop",
            "--symlink",
            "--python",
            f"{env.executable}",
        ],
        cwd=env.project_path,
        stdout=stdout,
        stderr=stderr,
    )



"""
Module responsible for interacting with local/remote projects.


Author: Tom Fleet
Created: 22/12/2021
"""

from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import Any

import humanize
import rtoml

from pytoil.api import API
from pytoil.config import Config
from pytoil.environments import Conda, Environment, Flit, Poetry, Requirements, Venv
from pytoil.exceptions import RepoNotFoundError


class Repo:
    def __init__(self, owner: str, name: str, local_path: Path) -> None:
        """
        Representation of a local/remote project.

        Args:
            owner (str): The owner (i.e. GitHub username)
            name (str): Project name.
            local_path (Path): The path this repo would have on the
                local filesystem, whether it currently exists or not.
        """
        self.owner = owner
        self.name = name
        self.local_path = local_path

    def __repr__(self) -> str:
        return (
            self.__class__.__qualname__
            + f"(owner={self.owner!r}, name={self.name!r},"
            f" local_path={self.local_path!r})"
        )

    __slots__ = ("owner", "name", "local_path")

    @property
    def clone_url(self) -> str:
        """
        The url to feed to `git clone` in order to
        clone this repo.
        """
        return f"https://github.com/{self.owner}/{self.name}.git"

    @property
    def html_url(self) -> str:
        """
        The url to the homepage of this repo.
        """
        return f"https://github.com/{self.owner}/{self.name}"

    @property
    def issues_url(self) -> str:
        """
        The url to the issues page of this repo.
        """
        return f"{self.html_url}/issues"

    @property
    def pulls_url(self) -> str:
        """
        The url to the pull requests page of this repo.
        """
        return f"{self.html_url}/pulls"

    def exists_local(self) -> bool:
        """
        Determines whether or not a `Repo` exists
        on the local filesystem.

        Returns:
            bool: True if exists locally, else False.
        """
        return self.local_path.exists()

    def exists_remote(self, api: API) -> bool:
        """
        Determines whether or not a `Repo` exists
        on GitHub.

        Args:
            api (API): The API object.

        Returns:
            bool: True if exists remote, else False.
        """
        return api.check_repo_exists(owner=self.owner, name=self.name)

    def _local_info(self) -> dict[str, Any] | None:  # pragma: no cover
        """
        Return local path information for the repo.
        """
        # Mostly just pathlib stuff here, not much point in us testing it
        # and doing so is a pain because we have to freeze time on the filesystem and in the test
        # if pathlib doesn't work we have bigger problems anyway
        try:
            return {
                "Name": self.local_path.name,
                "Created": humanize.naturaltime(
                    datetime.utcfromtimestamp(self.local_path.stat().st_birthtime),  # type: ignore[attr-defined]
                    when=datetime.utcnow(),
                ),
                "Updated": humanize.naturaltime(
                    datetime.utcfromtimestamp(self.local_path.stat().st_mtime),
                    when=datetime.utcnow(),
                ),
                "Local": True,
            }
        except FileNotFoundError:
            return None

    def _remote_info(self, api: API) -> dict[str, Any] | None:
        """
        Return remote API information for the repo.
        """
        return api.get_repo_info(self.name)

    def info(self, api: API) -> dict[str, Any]:
        """
        Return summary info about the repo in question.

        Prefers the API info as it is possible to get
        things like license and description.

        If local only will fall back to path.stat.

        Args:
            api (API): The API dependency.

        Raises:
            RepoNotFoundError: If repo not local or remote.

        Returns:
            Dict[str, Any]: Repository information.
        """
        info: dict[str, Any] = {}

        exists_remote = self.exists_remote(api=api)
        exists_local = self.exists_local()

        if exists_remote:
            if remote_info := self._remote_info(api=api):
                info.update(remote_info)
            # Might also exist locally
            info.update({"Local": exists_local})
        elif exists_local:
            if local_info := self._local_info():
                info.update(local_info)
            # We know it doesn't exist on GitHub if we got here
            info.update({"Remote": False})
        else:
            raise RepoNotFoundError(
                f"Repo: {self.name!r} does not exist locally or on GitHub."
            )

        return info

    def _file_exists(self, file: str) -> bool:
        """
        Convenience method to determine whether or not
        the repo root directory contains `file`.

        Args:
            file (str): Name of the file.

        Returns:
            bool: True if exists, else False.
        """
        return self.local_path.joinpath(file).exists()

    def is_setuptools(self) -> bool:
        """
        Is the project based on setuptools, i.e. does it
        have a `setup.cfg` or `setup.py`

        Returns:
            bool: True if setuptools, else False.
        """
        conditions = (
            self._file_exists("setup.cfg"),
            self._file_exists("setup.py"),
            self._specifies_build_tool("setuptools.build_meta"),
        )

        return any(conditions)

    def is_requirements(self) -> bool:
        """
        Is the project a python app with a requirements file?

        Returns:
            bool: True if yes, else False.
        """
        conditions = (
            self._file_exists("requirements.txt"),
            self._file_exists("requirements-dev.txt"),
        )

        return any(conditions)

    def has_pyproject_toml(self) -> bool:
        """
        Does the project have a `pyproject.toml` file.

        Returns:
            bool: True if yes, else False.
        """
        return self._file_exists("pyproject.toml")

    def is_conda(self) -> bool:
        """
        Is the project based on conda. The only reliable
        way of detecting this is looking for an `environment.yml`.

        Returns:
            bool: True if conda, else False.
        """
        return self._file_exists("environment.yml")

    def _specifies_build_tool(self, build_tool: str) -> bool:
        """
        Generalised method to check for a particular build tool
        in `pyproject.toml`

        Does more than a naive search for `build_tool`, actually checks the
        appropriate toml construction so if this method returns True, caller
        can be confident that `pyproject.toml` is valid.

        Args:
            build_tool (str): The build tool to check for e.g. `flit`, `poetry`.

        Returns:
            bool: True if `pyproject.toml` specifies that build tool, else False.
        """
        # If it doesn't have a pyproject.toml, bail early
        if not self.has_pyproject_toml():
            return False

        with open(self.local_path.joinpath("pyproject.toml")) as file:
            toml = rtoml.load(file)

        if build_system := toml.get("build-system"):
            if build_backend := build_system.get("build-backend"):
                return build_tool in build_backend.strip().lower()

        return False

    def is_pep621(self) -> bool:
        """
        Does the project comply with PEP 621.
        """
        if not self.has_pyproject_toml():
            return False

        with open(self.local_path.joinpath("pyproject.toml")) as file:
            toml = rtoml.load(file)

        if not toml.get("build-system"):
            return False

        if not toml.get("project"):
            return False

        return True

    def is_poetry(self) -> bool:
        """
        Does the project specify a poetry build system

        Returns:
            bool: True if yes, else False.
        """
        return self._specifies_build_tool("poetry")

    def is_flit(self) -> bool:
        """
        Does the project specify a flit build system.

        Returns:
            bool: True if yes, else False.
        """
        return self._specifies_build_tool("flit")

    def is_hatch(self) -> bool:
        """
        Does the project specify a hatch build system.

        Returns:
            bool: True if yes, else False.
        """
        return self._specifies_build_tool("hatchling.build")

    def dispatch_env(self, config: Config) -> Environment | None:
        """
        Returns the correct environment object for the calling `Repo`,
        or `None` if it cannot detect the environment.

        Therefore all usage should first check for `None`.

        Returns:
            Optional[Environment]: The correct environment object if it was
                able to detect, or `None`.
        """
        # This is where the magic happens for automatic environment detection
        # and installation

        # Each of the environment objects below implements the `Environment` Protocol
        # and has an `install_self` method that does the correct thing for it's environment

        if self.is_conda():
            return Conda(
                root=self.local_path, environment_name=self.name, conda=config.conda_bin
            )
        elif self.is_requirements():
            return Requirements(root=self.local_path)
        elif self.is_setuptools() or self.is_pep621():
            return Venv(root=self.local_path)
        elif self.is_poetry():
            return Poetry(root=self.local_path)
        elif self.is_flit():
            return Flit(root=self.local_path)
        else:
            # Could not autodetect, this is handled by the CLI
            return None



from __future__ import annotations

import subprocess
import sys
import tempfile
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.environments import Venv


def test_virtualenv():
    venv = Venv(root=Path("somewhere"))

    assert venv.project_path == Path("somewhere").resolve()
    assert venv.executable == Path("somewhere").resolve().joinpath(".venv/bin/python")
    assert venv.name == "venv"


def test_virtualenv_repr():
    venv = Venv(root=Path("somewhere"))
    assert repr(venv) == f"Venv(root={Path('somewhere')!r})"


@pytest.mark.parametrize(
    "exists_return, exists",
    [
        (True, True),
        (False, False),
    ],
)
def test_exists_returns_correct_value(
    mocker: MockerFixture, exists_return, exists: bool
):
    # Ensure Path.exists returns what we want it to
    mocker.patch(
        "pytoil.environments.virtualenv.Path.exists",
        autospec=True,
        return_value=exists_return,
    )

    venv = Venv(root=Path("somewhere"))
    assert venv.exists() is exists


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_calls_pip_correctly(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mock = mocker.patch(
        "pytoil.environments.virtualenv.subprocess.run",
        autospec=True,
    )

    venv = Venv(root=Path("somewhere"))

    venv.install(["black", "mypy", "isort", "flake8"], silent=silent)

    mock.assert_called_once_with(
        [
            f"{venv.executable}",
            "-m",
            "pip",
            "install",
            "black",
            "mypy",
            "isort",
            "flake8",
        ],
        cwd=venv.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_calls_pip_correctly(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mock = mocker.patch(
        "pytoil.environments.virtualenv.subprocess.run",
        autospec=True,
    )

    # Make it think there's already a venv
    mocker.patch(
        "pytoil.environments.virtualenv.Venv.exists",
        autospec=True,
        return_value=True,
    )

    venv = Venv(root=Path("somewhere"))

    venv.install_self(silent=silent)

    mock.assert_called_once_with(
        [f"{venv.executable}", "-m", "pip", "install", "-e", ".[dev]"],
        cwd=venv.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_creates_venv_if_not_one_already(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mock = mocker.patch(
        "pytoil.environments.virtualenv.subprocess.run",
        autospec=True,
    )

    # Make it think there isn't a venv
    mocker.patch(
        "pytoil.environments.virtualenv.Venv.exists",
        autospec=True,
        return_value=False,
    )

    # Mock out the venv.create method
    mock_create = mocker.patch(
        "pytoil.environments.virtualenv.Venv.create", autospec=True
    )

    venv = Venv(root=Path("somewhere"))

    venv.install_self(silent=silent)

    mock_create.assert_called_once()

    mock.assert_called_once_with(
        [f"{venv.executable}", "-m", "pip", "install", "-e", ".[dev]"],
        cwd=venv.project_path,
        stdout=stdout,
        stderr=stderr,
    )


def test_venv_create():
    with tempfile.TemporaryDirectory("w") as tmp:
        tmp_path = Path(tmp).resolve()
        venv = Venv(tmp_path)
        venv.create(silent=True)

        assert tmp_path.joinpath(".venv").exists()
        assert tmp_path.joinpath(".venv/pyvenv.cfg").exists()



"""
Styles for pytoil's output using rich.


Author: Tom Fleet
Created: 05/02/2022
"""

from __future__ import annotations

import sys

from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.style import Style
from rich.theme import Theme

__all__ = ("Printer", "printer")


class Printer:
    """
    Pytoil's default CLI output printer, designed for user
    friendly, colourful output, not for logging.
    """

    _pytoil_theme = Theme(
        styles={
            "title": Style(color="bright_cyan", bold=True),
            "info": Style(color="bright_cyan"),
            "warning": Style(color="yellow", bold=True),
            "error": Style(color="bright_red", bold=True),
            "error_message": Style(color="white", bold=True),
            "good": Style(color="bright_green"),
            "note": Style(color="white", bold=True),
            "subtle": Style(color="bright_black", italic=True),
        }
    )

    _pytoil_console = Console(theme=_pytoil_theme)

    __slots__ = ()

    def title(self, msg: str, spaced: bool = True) -> None:
        """
        Print a bold title message or section header.
        """
        to_print = f"{msg}"
        if spaced:
            to_print = f"{msg}\n"
        self._pytoil_console.print(to_print, style="title")

    def warn(self, msg: str, exits: int | None = None) -> None:
        """
        Print a warning message.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(f"⚠️  {msg}", style="warning")
        if exits is not None:
            sys.exit(exits)

    def info(self, msg: str, exits: int | None = None, spaced: bool = False) -> None:
        """
        Print an info message.

        If `exits` is not None, will call `sys.exit` with given code.

        If spaced is True, a new line will be printed before and after the message.
        """
        to_print = f"💡 {msg}"
        if spaced:
            to_print = f"\n💡 {msg}\n"

        self._pytoil_console.print(to_print, style="info")
        if exits is not None:
            sys.exit(exits)

    def sub_info(self, msg: str, exits: int | None = None) -> None:
        """
        Print a sub-info message.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(f"  ↪ {msg}")
        if exits is not None:
            sys.exit(exits)

    def error(self, msg: str, exits: int | None = None) -> None:
        """
        Print an error message.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(
            f"[error]✘  Error: [/error][error_message]{msg}[/error_message]"
        )
        if exits is not None:
            sys.exit(exits)

    def good(self, msg: str, exits: int | None = None) -> None:
        """
        Print a success message.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(f"✔  {msg}", style="good")
        if exits is not None:
            sys.exit(exits)

    def note(self, msg: str, exits: int | None = None) -> None:
        """
        Print a note, designed for supplementary info on another
        printer method.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(f"[note]Note:[/note] {msg}")
        if exits is not None:
            sys.exit(exits)

    def text(self, msg: str, exits: int | None = None) -> None:
        """
        Print default text.

        If `exits` is not None, will call `sys.exit` with given code.
        """
        self._pytoil_console.print(msg, style="default")
        if exits is not None:
            sys.exit(exits)

    def progress(self) -> Progress:
        """
        Return a pre-configured rich spinner.
        """
        text_column = TextColumn("{task.description}")
        spinner_column = SpinnerColumn("simpleDotsScrolling", style="bold white")
        return Progress(text_column, spinner_column, transient=True)

    def subtle(self, msg: str) -> None:
        """
        Print subtle greyed out text.
        """
        self._pytoil_console.print(msg, style="subtle", markup=None)


# Export a default printer
printer = Printer()



"""
Nox automation tasks for pytoil
"""

from __future__ import annotations

import json
import os
import shutil
import webbrowser
from pathlib import Path

import nox

# Nox config
nox.needs_version = ">=2022.1.7"
nox.options.error_on_external_run = True

# GitHub Actions
ON_CI = bool(os.getenv("CI"))

# Global project stuff
PROJECT_ROOT = Path(__file__).parent.resolve()
PROJECT_SRC = PROJECT_ROOT / "src"
PROJECT_TESTS = PROJECT_ROOT / "tests"

# Git info
DEFAULT_BRANCH = "main"

# VSCode
VSCODE_DIR = PROJECT_ROOT / ".vscode"
SETTINGS_JSON = VSCODE_DIR / "settings.json"

# Virtual environment stuff
VENV_DIR = PROJECT_ROOT / ".venv"
PYTHON = os.fsdecode(VENV_DIR / "bin" / "python")

# Python to use for non-test sessions
DEFAULT_PYTHON: str = "3.11"

# All supported python versions for pytoil
PYTHON_VERSIONS: list[str] = [
    "3.9",
    "3.10",
    "3.11",
]

# "dev" should only be run if no virtual environment found and we're not on CI
# i.e. someone is using nox to set up their local dev environment to
# work on pytoil
if not VENV_DIR.exists() and not ON_CI:
    nox.options.sessions = ["dev"]
else:
    nox.options.sessions = ["test", "coverage", "lint", "docs"]


@nox.session(python=DEFAULT_PYTHON)
def dev(session: nox.Session) -> None:
    """
    Sets up a python dev environment for the project if one doesn't already exist.

    This session will:
    - Create a python virtualenv for the session
    - Install the `virtualenv` cli tool into this environment
    - Use `virtualenv` to create a global project virtual environment
    - Invoke the python interpreter from the global project environment to install
      the project and all it's development dependencies.
    """
    # Check if dev has been run before
    # this prevents manual running nox -s dev more than once
    # thus potentially corrupting an environment
    if VENV_DIR.exists():
        session.error(
            "There is already a virtual environment deactivate and remove it "
            "before running 'dev' again"
        )

    # Create the project virtual environment using virtualenv
    # installed into this sessions virtual environment
    # confusing but it works!
    session.install("virtualenv")
    session.run("virtualenv", os.fsdecode(VENV_DIR), silent=True)

    # Use the venv's interpreter to install the project along with
    # all it's dev dependencies, this ensure it's installed
    # in the right way
    session.run(
        PYTHON,
        "-m",
        "pip",
        "install",
        "--upgrade",
        "pip",
        "setuptools",
        "wheel",
        silent=True,
        external=True,
    )
    session.run(PYTHON, "-m", "pip", "install", "-e", ".[dev]", external=True)

    if bool(shutil.which("code")) or bool(shutil.which("code-insiders")):
        # Only do this is user has VSCode installed
        set_up_vscode(session)


@nox.session(python=PYTHON_VERSIONS)
def test(session: nox.Session) -> None:
    """
    Runs the test suite against all supported python versions.
    """
    session.install("--upgrade", "pip", "setuptools", "wheel")
    # Tests require the package to be installed
    session.install(".")
    session.install(
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "pytest-httpx",
        "pytest-randomly",
        "covdefaults",
        "freezegun",
    )

    session.run("pytest", f"--cov={PROJECT_SRC}", f"{PROJECT_TESTS}")


@nox.session(python=DEFAULT_PYTHON)
def codecov(session: nox.Session) -> None:
    """
    Generate a codecov xml report for CI.
    """
    session.install("--upgrade", "pip", "setuptools", "wheel")

    session.install(".")
    session.install(
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "pytest-httpx",
        "pytest-randomly",
        "covdefaults",
        "freezegun",
    )

    session.run("pytest", f"--cov={PROJECT_SRC}", f"{PROJECT_TESTS}")
    session.run("coverage", "xml")


@nox.session(python=DEFAULT_PYTHON)
def coverage(session: nox.Session) -> None:
    """
    Test coverage analysis.
    """
    session.install("--upgrade", "pip", "setuptools", "wheel")
    session.install("coverage[toml]", "covdefaults")

    session.run("coverage", "report", "--show-missing")


@nox.session(python=DEFAULT_PYTHON)
def lint(session: nox.Session) -> None:
    """
    Run pre-commit linting.
    """
    session.install("--upgrade", "pip", "setuptools", "wheel")
    session.install("pre-commit")

    session.run("pre-commit", "run", "--all-files")


@nox.session(python=DEFAULT_PYTHON)
def docs(session: nox.Session) -> None:
    """
    Builds the project documentation. Use '-- serve' to see changes live.
    """
    session.install("--upgrade", "pip", "setuptools", "wheel")
    session.install("mkdocs", "mkdocs-material")

    if "serve" in session.posargs:
        webbrowser.open(url="http://127.0.0.1:8000/pytoil/")
        session.run("mkdocs", "serve")
    else:
        session.run("mkdocs", "build", "--clean")


@nox.session
def deploy_docs(session: nox.Session) -> None:
    """
    Used by GitHub actions to deploy docs to GitHub Pages.
    """
    if not (token := os.getenv("GITHUB_TOKEN")):
        session.error("Cannot deploy docs without a $GITHUB_TOKEN environment variable")

    session.install("--upgrade", "pip", "setuptools", "wheel")
    session.install("mkdocs", "mkdocs-material")

    if ON_CI:
        session.run(
            "git",
            "remote",
            "add",
            "gh-token",
            f"https://{token}@github.com/FollowTheProcess/pytoil.git",
            external=True,
        )
        session.run("git", "fetch", "gh-token", external=True)
        session.run("git", "fetch", "gh-token", "gh-pages:gh-pages", external=True)

        session.run("mkdocs", "gh-deploy", "-v", "--clean", "--remote-name", "gh-token")
    else:
        session.run("mkdocs", "gh-deploy")


@nox.session(python=DEFAULT_PYTHON)
def build(session: nox.Session) -> None:
    """
    Builds the package sdist and wheel.
    """
    session.install("build")
    session.run("python", "-m", "build", ".")


def set_up_vscode(session: nox.Session) -> None:
    """
    Helper function that will set VSCode's workspace settings
    to use the auto-created virtual environment and enable
    pytest support.

    If called, this function will only do anything if
    there aren't already VSCode workspace settings defined.

    Args:
        session (nox.Session): The enclosing nox session.
    """
    if not VSCODE_DIR.exists():
        session.log("Setting up VSCode Workspace.")
        VSCODE_DIR.mkdir(parents=True)
        SETTINGS_JSON.touch()

        settings = {
            "python.defaultInterpreterPath": PYTHON,
            "python.testing.pytestEnabled": True,
            "python.testing.pytestArgs": [PROJECT_TESTS.name],
        }

        with open(SETTINGS_JSON, mode="w", encoding="utf-8") as f:
            json.dump(settings, f, sort_keys=True, indent=4)



from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.environments import Requirements


def test_requirements():
    env = Requirements(root=Path("somewhere"))

    assert env.project_path == Path("somewhere").resolve()
    assert env.name == "requirements file"
    assert env.executable == Path("somewhere").resolve().joinpath(".venv/bin/python")


def test_requirements_repr():
    env = Requirements(root=Path("somewhere"))
    assert repr(env) == f"Requirements(root={Path('somewhere')!r})"


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_venv_exists(mocker: MockerFixture, silent: bool, stdout, stderr):
    mocker.patch(
        "pytoil.environments.reqs.Requirements.exists",
        autospec=True,
        return_value=True,
    )

    mock = mocker.patch("pytoil.environments.reqs.subprocess.run", autospec=True)

    env = Requirements(root=Path("somewhere"))

    env.install_self(silent=silent)

    mock.assert_called_once_with(
        [f"{env.executable}", "-m", "pip", "install", "-r", "requirements.txt"],
        cwd=env.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_venv_doesnt_exist(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mocker.patch(
        "pytoil.environments.reqs.Requirements.exists",
        autospec=True,
        return_value=False,
    )

    mock = mocker.patch("pytoil.environments.reqs.subprocess.run", autospec=True)

    mock_create = mocker.patch(
        "pytoil.environments.reqs.Requirements.create", autospec=True
    )

    env = Requirements(root=Path("somewhere"))

    env.install_self(silent=silent)

    mock_create.assert_called_once()

    mock.assert_called_once_with(
        [f"{env.executable}", "-m", "pip", "install", "-r", "requirements.txt"],
        cwd=env.project_path,
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_install_self_requirements_dev(
    mocker: MockerFixture, silent: bool, stdout, stderr
):
    mocker.patch(
        "pytoil.environments.reqs.Requirements.exists",
        autospec=True,
        return_value=True,
    )

    # Make it think that requirements-dev.txt exists
    mocker.patch(
        "pytoil.environments.reqs.Path.exists",
        autospec=True,
        return_value=True,
    )

    mock = mocker.patch("pytoil.environments.reqs.subprocess.run", autospec=True)

    env = Requirements(root=Path("somewhere"))

    env.install_self(silent=silent)

    mock.assert_called_once_with(
        [f"{env.executable}", "-m", "pip", "install", "-r", "requirements-dev.txt"],
        cwd=env.project_path,
        stdout=stdout,
        stderr=stderr,
    )



"""
Global defaults for pytoil.

Author: Tom Fleet
Created: 21/12/2021
"""

from __future__ import annotations

import os
from pathlib import Path

# Default path for pytoil's config file
CONFIG_FILE: Path = Path.home().joinpath(".pytoil.toml").resolve()

# Valid pytoil config keys
CONFIG_KEYS: set[str] = {
    "projects_dir",
    "token",
    "username",
    "editor",
    "conda_bin",
    "common_packages",
    "git",
}

# Pytoil meta stuff
PYTOIL_DOCS_URL: str = "https://followtheprocess.github.io/pytoil/"
PYTOIL_ISSUES_URL: str = "https://github.com/FollowTheProcess/pytoil/issues"

# Defaults for pytoil config
PROJECTS_DIR: Path = Path.home().joinpath("Development").resolve()
TOKEN: str = os.getenv("GITHUB_TOKEN", "")
USERNAME: str = ""
EDITOR: str = os.getenv("EDITOR", "")
CONDA_BIN: str = "conda"
COMMON_PACKAGES: list[str] = []
GIT: bool = True

# Config Schema
CONFIG_SCHEMA = """

# The .pytoil.toml config file

## projects_dir *(str)*

The absolute path to where you keep your development projects
(e.g. /Users/you/Projects).

## token *(str)*

Your GitHub personal access token. This must have a minimum of repo read access. See the documentation here:
https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/creating-a-personal-access-token

Pytoil will try and get this from the config file initially, then fall back to the $GITHUB_TOKEN environment
variable. If neither of these places are set, you will not be able to use pytoil commands that rely on the
GitHub API. Pytoil will notify you of this when any of these commands are called.

## username *(str)*

Your GitHub username. Pytoil needs this so it can construct urls to your projects.

## editor *(str)*

The name of the editor you'd like pytoil to use to open projects. Note: This editor must
be directory-aware and have a command line binary that can be used to launch projects.

For example:
    `code /path/to/my/project`
    `code-insiders /path/to/my/project`
    `pycharm /path/to/my/project`

If the key is not present in the config file, pytoil will fall back to $EDITOR, which may fail
if the configured $EDITOR is not directory-aware e.g. things like vim and nvim.

If the key is set to the literal string "None" (case-insensitive), pytoil will not attempt to open
projects for you.

Otherwise, the value of `editor` will be used as the name of the command line binary used to open
projects e.g. `code, `code-insiders`, `pycharm` etc.


## conda_bin *(str)*

The name of the binary to use when performing conda operations. Either "conda" (default)
or "mamba"

## common_packages *(List[str])*

A list of python packages to inject into every virtual environment pytoil creates
(e.g. linters, formatters and other dev dependencies).

Any versioning syntax (e.g. mypy>=0.902) will work as expected here as these packages
are passed straight through to installation tools like pip and conda.

## git *(bool)*

Whether or not you want pytoil to create an empty git repo when you make a new project with
'pytoil new'. This can also be disabled on a per use basis using the '--no-git' flag.
"""



"""
Module responsible for handling python environments
with a `requirements.txt` (or `requirements-dev.txt`)


Author: Tom Fleet
Created: 26/12/2021
"""

from __future__ import annotations

import subprocess
import sys
from pathlib import Path

from pytoil.environments.virtualenv import Venv


class Requirements(Venv):
    def __init__(self, root: Path) -> None:
        self.root = root
        super().__init__(root)

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(root={self.root!r})"

    __slots__ = ("root",)

    @property
    def name(self) -> str:
        return "requirements file"

    def install_self(self, silent: bool = False) -> None:
        """
        Installs everything in the requirements file into
        a python environment.
        """
        if not self.exists():
            self.create(silent=silent)

        requirements_file = "requirements.txt"

        if self.project_path.joinpath("requirements-dev.txt").exists():
            requirements_file = "requirements-dev.txt"

        subprocess.run(
            [f"{self.executable}", "-m", "pip", "install", "-r", requirements_file],
            cwd=self.project_path,
            stdout=subprocess.DEVNULL if silent else sys.stdout,
            stderr=subprocess.DEVNULL if silent else sys.stderr,
        )



from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from pytoil.exceptions import GitNotInstalledError
from pytoil.git import Git


def test_git_instanciation_default():
    git = Git()

    assert git.git == shutil.which("git")


def test_git_instanciation_passed():
    git = Git(git="/some/path/to/git")

    assert git.git == "/some/path/to/git"


def test_git_repr():
    git = Git(git="hellogit")

    assert repr(git) == "Git(git='hellogit')"


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_git_init(mocker: MockerFixture, silent: bool, stdout, stderr):
    mock = mocker.patch("pytoil.git.git.subprocess.run", autospec=True)

    git = Git(git="notgit")

    git.init(Path("somewhere"), silent=silent)

    mock.assert_called_once_with(
        ["notgit", "init"],
        cwd=Path("somewhere"),
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_git_clone(mocker: MockerFixture, silent: bool, stdout, stderr):
    mock = mocker.patch("pytoil.git.git.subprocess.run", autospec=True)

    git = Git(git="notgit")

    git.clone(
        url="https://nothub.com/some/project.git", cwd=Path("somewhere"), silent=silent
    )

    mock.assert_called_once_with(
        ["notgit", "clone", "https://nothub.com/some/project.git"],
        cwd=Path("somewhere"),
        stdout=stdout,
        stderr=stderr,
    )


def test_instantiation_raises_if_git_not_insalled():
    with pytest.raises(GitNotInstalledError):
        Git(git=None)


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_git_set_upstream(mocker: MockerFixture, silent: bool, stdout, stderr):
    mock = mocker.patch("pytoil.git.git.subprocess.run", autospec=True)

    git = Git(git="notgit")

    git.set_upstream(owner="me", repo="project", cwd=Path("somewhere"), silent=silent)

    mock.assert_called_once_with(
        ["notgit", "remote", "add", "upstream", "https://github.com/me/project.git"],
        cwd=Path("somewhere"),
        stdout=stdout,
        stderr=stderr,
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_git_add_all(mocker: MockerFixture, silent: bool, stdout, stderr):
    mock = mocker.patch("pytoil.git.git.subprocess.run", autospec=True)

    git = Git(git="notgit")

    git.add(cwd=Path("somewhere"), silent=silent)

    mock.assert_called_once_with(
        ["notgit", "add", "-A"], cwd=Path("somewhere"), stdout=stdout, stderr=stderr
    )


@pytest.mark.parametrize(
    "silent, stdout, stderr",
    [
        (True, subprocess.DEVNULL, subprocess.DEVNULL),
        (False, sys.stdout, sys.stderr),
    ],
)
def test_git_commit(mocker: MockerFixture, silent: bool, stdout, stderr):
    mock = mocker.patch("pytoil.git.git.subprocess.run", autospec=True)

    git = Git(git="notgit")

    git.commit(message="Commit message", cwd=Path("somewhere"), silent=silent)

    mock.assert_called_once_with(
        ["notgit", "commit", "-m", "Commit message"],
        cwd=Path("somewhere"),
        stdout=stdout,
        stderr=stderr,
    )
